/* 
 * Copyright (c) 1990, 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *	Utah $Hdr: grfreg.h 1.12 94/12/16$
 */

/* 300 bitmapped display hardware primary id */
#define GRFHWID		0x39

/* 300 internal bitmapped display address */
#define GRFIADDR	0x560000

/* 300 hardware secondary ids */
#define GID_GATORBOX	1
#define	GID_TOPCAT	2
#define GID_RENAISSANCE	4
#define GID_LRCATSEYE	5
#define GID_HRCCATSEYE	6
#define GID_HRMCATSEYE	7
#define GID_DAVINCI	8
#define GID_XXXCATSEYE	9
#define GID_XGENESIS   11
#define GID_TIGER      12
#define GID_YGENESIS   13
#define GID_HYPERION   14

/* 800 hardware secondary ids */
#define GID_FIREEYE	6

/* 700 does not need a secondary id, but we want it to fit the model */
#define GID_STI		69

#ifdef hp300
typedef unsigned char	grftype;
#else
typedef unsigned long	grftype;
#endif

struct	grfreg {
	grftype	gr_pad0,
		gr_id,		/* +0x01 */
		gr_pad1[0x3],
		gr_fbwidth_h,	/* +0x05 */
		gr_pad2,
		gr_fbwidth_l,	/* +0x07 */
		gr_pad3,
		gr_fbheight_h,	/* +0x09 */
		gr_pad4,
		gr_fbheight_l,	/* +0x0B */
		gr_pad5,
		gr_dwidth_h,	/* +0x0D */
		gr_pad6,
		gr_dwidth_l,	/* +0x0F */
		gr_pad7,
		gr_dheight_h,	/* +0x11 */
		gr_pad8,
		gr_dheight_l,	/* +0x13 */
		gr_pad9,
		gr_id2,		/* +0x15 */
		gr_pad10[0x47],
		gr_fbomsb,	/* +0x5d */
		gr_pad11,
		gr_fbolsb;	/* +0x5f */
};

#ifndef hp300
#ifdef hp700
/*
 * 700 specific stuff
 */
#ifdef MACH_KERNEL
#include <grf_stireg.h>
#else
#include <hpsgc/grf_stireg.h>
#endif

/*
 * We need to keep track of the ROM address and the STI info.
 * Use this table, indexed by unit number.
 */
struct grfdev {
	caddr_t		   romaddr;		/* ROM address */
	caddr_t		   hpa;			/* device address */
	int		   type;		/* STI_TYPE_* */
	struct sti_entry   *ep;			/* STI routines */
	struct sti_region  regions[STI_REGIONS];/* Region list */
	struct sti_config  config;		/* "global" config info */
	char 		   devname[STI_NAMELEN];/* Name for GCDESCRIBE */
};

#else

/*
 * 800 specific stuff
 */

/*
 * We need to keep track of the HPAs for each set of cards.
 * Use this table, indexed by unit number.
 */
struct grfdev {
	struct iomod *fb_hpa;		/* frame buffer HPA */
	struct iomod *cs_hpa;		/* control space HPA */
	int          *fb_spa;		/* frame buffer SPA */
	int          *cs_spa;		/* control space SPA */
	u_int	     eim;		/* external interrupt mask */
};

/* HPA status bits. */
#define G0_RY	    0x40		/*  Ready Bit */
#define G0_FE	    0x80		/*  Fatal Error */
#define G0_HE	    0x100		/*  Hard Error */
#define G0_LOST	    0x10000		/*  Power Lost */
#define G0_GINT	    0x10000000
#define G0_INT1     0x20000000
#define G0_INT7	    0x40000000
#define G0_NMI	    0x40000000
#define G0_TIMEDOUT 0x80000000

/* HPA I/O commands */
#define CMD_CLEAR 3
#define CMD_RESET 5

/* The maximum number of microseconds it takes to perform a CMD_RESET */
#define MAX_RESET_TIME	5

/*
 * See grf_rbreg.h.
 * Offset of the HIL registers from the start of the control space.
 */
#define HILOFFSET 0x5004
#endif
#endif
