/* 
 * Copyright (c) 1991, 1992, 1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: glue.c 1.30 95/04/26$
 */
/*
 * Just what the name says
 */

#ifndef MACH_KERNEL
#include <sys/cdefs.h>
#endif
#include <sys/types.h>
#include <machine/asp.h>
#ifdef MACH_KERNEL
#include <machine/machspl.h>
#include <mach/machine/vm_types.h>
#include <mach_glue.h>
#include <device.h>
#else
#include <machine/spl.h>
#include <sys/param.h>
#include <hpsgc/bsd_glue.h>
#include <hpdev/device.h>
#endif

#include "dca.h"
#if NDCA > 0
#ifdef MACH_KERNEL
#include <dcareg.h>
#else
#include <hpdev/dcareg.h>
#endif
#endif

#include "hil.h"
#if NHIL > 0
#ifdef MACH_KERNEL
#define	SPLPOWER INTPRI_01
#include <hilreg.h>
#include <hilioctl.h>
#include <hilvar.h>
#else
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <hpdev/hilreg.h>
#include <hpdev/hilioctl.h>
#include <hpdev/hilvar.h>
#endif
#endif

#include "lan.h"
#include "scsiha.h"
#include "sd.h"
#include "st.h"
#include "gkd.h"

#ifndef MACH_KERNEL
#include "eisa.h"
#if NEISA > 0
#include <hpsgc/atbus.h>
extern int eisa_irqs[];
extern int eisa_scno[];
extern int eisa_msiz;
#endif

#if NSCSIHA > 0
#include <hpsgc/scsi.h>

#if NSD > 0
extern struct scsi_device scsi_dasd[];
#endif

#if NST > 0
extern struct scsi_device scsi_sequential[];
#endif
#endif
#endif	/* !MACH_KERNEL */

#include "medusa.h"
#include "grf.h"
#if NGRF > 0
#include <machine/iomod.h>
#endif

#if defined(DEBUG) && !defined(PANICBUTTON) && !defined(MACH_KERNEL)
#define PANICBUTTON	1	/* XXX */
#endif

#ifndef	MACH_KERNEL
u_long	dcache_line_size = 32;			/* XXX for scsi */
#endif
u_long	scsi_clock_freq = 0;			/* XXX for scsi */
int	isgecko = 0;				/* XXX */

ioconf()
{
	struct hp_device hd;
	struct modtab *m, *getmod();

	if (isgecko) {
		lasiinit();
		waxinit();
	} else
		aspinit();

#if NDCA > 0
	if (isgecko) {
		m = getmod(IODC_TP_FIO, SVMOD_FIO_GRS232, 0);
		if (m) {
			extern int dcaintr();

			hd.hp_addr = (char *)m->m_hpa;
			hd.hp_unit = 0;
			hd.hp_flags = (1 << hd.hp_unit);
			dcaprobe(&hd);
			aspitab(INT_RS232_1, SPLTTY, dcaintr, 0, 0);
			/* XXX assumes second port is on WAX */
			m = getmod(IODC_TP_FIO, SVMOD_FIO_GRS232, 1);
			if (m && m->m_hpa == 0xf0202000) {
				hd.hp_addr = (char *)m->m_hpa;
				hd.hp_unit = 1;
				hd.hp_flags = (1 << hd.hp_unit);
				dcaprobe(&hd);
				waxitab(0x40, SPLTTY, dcaintr, 1, 0);
			}
		}
	} else {
		m = getmod(IODC_TP_FIO, SVMOD_FIO_RS232, 0);
		if (m) {
			extern int dcaintr();

			hd.hp_addr = (char *)m->m_hpa;
			hd.hp_unit = (m->m_fixed == 4) ? 0 : 1;
			hd.hp_flags = (1 << hd.hp_unit);
			dcaprobe(&hd);
			aspitab(hd.hp_unit ? INT_RS232_2 : INT_RS232_1,
				SPLTTY, dcaintr, hd.hp_unit, 0);
			m = getmod(IODC_TP_FIO, SVMOD_FIO_RS232, 1);
			if (m) {
				hd.hp_addr = (char *)m->m_hpa;
				hd.hp_unit = (m->m_fixed == 4) ? 0 : 1;
				hd.hp_flags = (1 << hd.hp_unit);
				dcaprobe(&hd);
				aspitab(hd.hp_unit ? INT_RS232_2 : INT_RS232_1,
					SPLTTY, dcaintr, hd.hp_unit, 0);
			}
		}
	}
#endif
#if NLAN > 0
	m = getmod(IODC_TP_FIO, isgecko ? SVMOD_FIO_GLAN : SVMOD_FIO_LAN, 0);
	if (m) {
		hd.hp_addr = (char *)m->m_hpa;
		hd.hp_unit = 0;
#ifdef DEBUG
		if (m->m_lanaddr[0] == 0)
			printf("WARNING: could not get LAN addr from PDC, trying standard locations\n");
#endif
		hd.hp_ctlr = (int)m->m_lanaddr;	/* XXX */
		pc586probe(&hd);
		pc586attach(&hd);
	}
#endif
#if NGKD > 0
	m = getmod(IODC_TP_FIO, SVMOD_FIO_GPCIO, 0);
	if (m) {
		hd.hp_addr = (char *)m->m_hpa;
		hd.hp_unit = 0;
		if (gkdprobe(&hd))
			gkdattach(&hd);
	}
#endif /* NGKD > 0 */
#if NMEDUSA > 0
	medconfig();
#endif
#if NSCSIHA > 0
	m = getmod(IODC_TP_FIO, isgecko ? SVMOD_FIO_GSCSI : SVMOD_FIO_SCSI, 0);
	if (m) {
		struct hp_ctlrinfo scsictlr;
		struct hp_businfo scsibus;
#ifdef DEBUG
		extern struct asp_ctrl *CORE_CTRL;
		extern int itick_per_tick, hz;
		u_long freq;

		/*
		 * Determine the SCSI clock frequency
		 * XXX we use a magic value out of the PDC now.
		 */
		freq = itick_per_tick * hz;
		if (CORE_CTRL->iostat.scsi_sel == 0)
			freq >>= 1;
		if (m->m_scsiclk && m->m_scsiclk != freq)
			printf("WARNING: SCSI HW freq (%d) != computed (%d), using HW\n",
			       m->m_scsiclk, freq);
#endif
		/*
		 * XXX avoid changing the SCSI driver for now.
		 * Set the global scsi_clock_freq to reflect the value
		 * for the current controller.
		 */
		scsi_clock_freq = m->m_scsiclk;
		if (scsi_clock_freq == 0)
			scsi_clock_freq = 50000000;

		scsictlr.ctlrnum = 0;
		scsictlr.parent_ctlrinfop = &scsictlr;
		scsictlr.ctlr_businfop = &scsibus;
		scsictlr.numcsrs = 1;
		scsictlr.vcsrs[0] = (vm_offset_t)m->m_hpa;
		scsibus.bustype = BUSTYPE_SYS;
		if (ncr53c700_probe(&scsictlr) == 0)
			panic("ioconf: internal SCSI not found");
		printf("scsi%d at sysbus0\n", scsictlr.ctlrnum);
#ifndef	MACH_KERNEL
		scsi_probe(&scsictlr);
#endif
	}
#endif
#if NSCSIHA > 1
	m = getmod(IODC_TP_FIO, SVMOD_FIO_FWSCSI, 0);
	if (m) {
		struct hp_ctlrinfo scsictlr;
		struct hp_businfo scsibus;

		scsictlr.ctlrnum = 1;
		scsictlr.parent_ctlrinfop = &scsictlr;
		scsictlr.ctlr_businfop = &scsibus;
		scsictlr.numcsrs = 1;
		scsictlr.vcsrs[0] = (vm_offset_t)m->m_hpa;
		scsibus.bustype = BUSTYPE_SYS;
		if (ncr53c700_probe(&scsictlr)) {
			printf("scsi%d at sysbus0\n", scsictlr.ctlrnum);
#ifndef	MACH_KERNEL
			scsi_probe(&scsictlr);
#endif
		}
	}
#endif
#if NEISA > 0
	m = getmod(IODC_TP_BA, SVMOD_BA_EISA, 0);
	if (m) {
		struct hp_ctlrinfo eisactlr, scsictlr;
		struct hp_businfo eisabus;
		int slot;

		eisactlr.ctlrnum = 0;
		eisactlr.parent_ctlrinfop = &eisactlr;
		eisactlr.ctlr_businfop = &eisabus;
		eisactlr.numcsrs = 1;
		eisactlr.csrs[0] = eisactlr.vcsrs[0] = (vm_offset_t)m->m_hpa;
		eisabus.bustype = BUSTYPE_EISA;
		atbus_probe(&eisactlr);
		printf("eisa0 at sysbus0\n");
#if NSCSIHA > 2
		scsictlr.parent_ctlrinfop = &eisactlr;
		scsictlr.ctlr_businfop = &eisabus;
		scsictlr.numcsrs = 1;
		scsictlr.csrs[0] = scsictlr.vcsrs[0] =
			eisactlr.csrs[0] + 0x1000;
		for (slot = 1; slot < eisa_msiz; slot++) {
			scsictlr.irq = eisa_irqs[slot];
			scsictlr.ctlrnum = eisa_scno[slot];
			if (scsictlr.ctlrnum >= 0 &&
			    ncr53c700_probe(&scsictlr)) {
				printf("scsi%d at eisa0 slot %d irq %d\n",
				       scsictlr.ctlrnum, slot-1, scsictlr.irq);
				/*
				 * We seem to need a delay here or else
				 * the initial TEST UNIT READY command in
				 * sdinit() fails.
				 */
				delay(1000000);
				atbus_resolver(RESOLVER_ADD_HANDLER,
					       eisactlr.ctlrnum, scsictlr.irq,
					       ATBUS_RESOLVER_LEVEL_INT);
				scsi_probe(&scsictlr);
			}
			scsictlr.csrs[0] += 0x1000;
			scsictlr.vcsrs[0] += 0x1000;
		}
#endif
		printf("eisa0 configuration done\n");
	}
#endif
#if NGRF > 0
	grfconfig();
#endif
#if NHIL > 0
	m = getmod(IODC_TP_FIO, SVMOD_FIO_HIL, 0);
	if (m) {
		extern int hilspl;
		extern int hilint(), nmihand();

		hilspl = SPLHIL;
		/*
		 * Under the ASP we get an initial interrupt when we unmask
		 * the device for the first time.  Hence, we must have the
		 * appropriate software structures initialized before we do
		 * the aspitab call (which in turn must be done before
		 * calling hilinit).
		 */
		hilsoftinit(0, (struct hil_dev *)m->m_hpa);
		if (isgecko)
			waxitab(0x2, SPLHIL, hilint, 0, 0);
		else
			aspitab(INT_8042_GEN, SPLHIL, hilint, 0, 0);
		hilinit(0, (struct hil_dev *)m->m_hpa);
#ifdef PANICBUTTON
		/*
		 * XXX not really proper to set SPL above SPLASP for
		 * this, but it works ok in this instance.
		 */
		if (isgecko)
			waxitab(0x04, SPLPOWER, nmihand, 0, 0);
		else
			aspitab(INT_8042_HI, SPLPOWER, nmihand, 0, 0);
#endif
	}
#endif
}

#if NSCSIHA > 0
#ifndef	MACH_KERNEL
scsi_probe(scsictlrp)
	struct hp_ctlrinfo *scsictlrp;
{
	static int first = 1;
	int ctlr, unit;
	dev_t dev;
	extern int dkn;
#ifdef DEBUG
	int sper, wid;
#endif

	ctlr = scsictlrp->ctlrnum;
#if NSD > 0
	if (first) {
		for (unit = 0; unit < NSD; unit++)
			scsi_dasd[unit].unused = -1;
		first = 0;
	}
	dev = makedev(7, 0);
	ncr53c700_timeout(scsictlrp, 1);
	for (unit = 0; unit < NSD; unit++) {
		if (scsi_dasd[unit].bus == ctlr && sdinit(dev, 1) == 0) {
			printf("sd%d at scsi%d, slave %d\n",
			       unit, ctlr, scsi_dasd[unit].target);
#ifdef DEBUG
			ncr53c700_params(ctlr, scsi_dasd[unit].target,
					 &sper, &wid);
			printf("sd%d: ", unit);
			if (sper == -1)
				printf("async");
			else {
				if (wid > 8)
					sper /= (wid / 8);
				printf("sync (%d.%d MB/sec max)",
				       1000 / sper, (10000 / sper) % 10);
			}
			printf(", %d-bit data transfers\n", wid);
#endif
			scsi_dasd[unit].unused = dkn-1;
		}
		dev += 8;
	}
	ncr53c700_timeout(scsictlrp, 0);
#endif
#if NST > 0
	dev = makedev(20, 0);
	ncr53c700_timeout(scsictlrp, 1);
	for (unit = 0; unit < NST; unit++) {
		if (scsi_sequential[unit].bus == 0 && stinit(dev) == 0)
			printf("st%d at scsi%d, slave %d\n",
			       unit, ctlr, scsi_sequential[unit].target);
		dev++;
	}
	ncr53c700_timeout(scsictlrp, 0);
#endif
}
#endif
#endif

#ifdef PANICBUTTON
int panicbutton = PANICBUTTON;
#if NHIL > 0
int crashandburn = 0;
int candbdelay = 50;		/* 1/HZths--give em half a second */

candbtimer()
{
	crashandburn = 0;
}

nmihand(unit)
	int unit;
{
	if (kbdnmi(unit)) {
		printf("Got a keyboard NMI\n");
		if (panicbutton) {
			if (crashandburn) {
				extern char *panicstr;

				crashandburn = 0;
				panic(panicstr ?
				      "forced crash, nosync" : "forced crash");
			}
			crashandburn++;
			timeout(candbtimer, (caddr_t)0, candbdelay);
		}
		return;
	}
	/* panic?? */
	printf("unexpected NMI interrupt ignored\n");
}
#endif
#endif
