/* 
 * Copyright (c) 1991-1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: kgdb.c 1.11 94/12/14$
 */

/*
 *	Machine dependent part of kgdb for parisc
 */

#ifdef KGDB

#include <mach/vm_prot.h>
#include <mach/thread_status.h>
#include <mach/vm_param.h>
#include <kern/task.h>
#include <kern/thread.h>
#include <kern/processor.h>
#include <kern/current.h>
#include <vm/vm_map.h>
#include <machine/trap.h>
#include <machine/psl.h>

#ifndef KGDBDEV
#define KGDBDEV		-1
#endif
#ifndef KGDBRATE
#define KGDBRATE	9600
#endif

#ifndef	CMU_PMAP
#define kernel_equiv_end	equiv_end
#define kernel_text_end		text_end
#endif	/* CMU_PMAP */


/*
 *	flag in pcb to indicate that we have saved kgdb state 
 */
#define	KGDB_STEP	2

#ifdef KGDB_LAN
#include <sys/types.h>
#include <machine/asp.h>
#include <device.h>
extern boolean_t kgdb_initialized;
#define kgdb_enabled	kgdb_initialized
#define kgdb_debug	0
#else
extern int		kgdb_debug;
extern boolean_t	kgdb_enabled;

/*
 * Device number we are using for kgdb.
 * If we ever see a '!' character in this unit then we drop into kgdb.
 */
int	kgdb_dev = KGDBDEV;

/*
 * Baud rate to use for kgdb serial connection.
 */
int	kgdb_rate = KGDBRATE;
#endif

/*
 *	These are locations in the kernel text used for debugging:
 *
 *	user_sstep is a kgdb break instruction on the gateway page
 *	user_nullify is the next instruction after it
 *	kgdb_breaK_inst is the address of the break in kgdb_break
 */
extern void user_sstep();
extern void user_nullify();
extern void kgdb_break_inst();

/*
 *	values from the pmap module
 */
extern vm_offset_t kernel_equiv_end;
extern vm_offset_t kernel_text_end;

/*
 *	kernel map
 */
extern vm_map_t	kernel_map;


#ifdef KGDB_LAN
/*
 *	Initialize the network stuff.
 */
void kgdb_net_init()
{
	struct hp_device hd;

#if 0
	hd.hp_addr = (char *)CORE_LAN;
	hd.hp_unit = 0;
	pc586probe(&hd);
	pc586attach(&hd);
#endif
	if (!pc586init(0)) {
		printf("first init failed, try again...\n");
		delay(1000000);
		(void) pc586init(0);
	}
}	

/*
 *	Check for incoming network packet.
 */
void kgdb_net_intr()
{
	pc586intr(0);
}

/*
 *	Transmit network packet.
 */
void kgdb_net_xmt(buf, len)
	char	*buf;
	int	len;
{
	pc586debug_xmt(0, buf, len);
}	
#else
/*
 *	Get a character from the kgdb port, return 0 if timeout
 */
char kgdb_getc()
{
	return dcacngetc(1, TRUE);
}

/*
 *	Put a character to the kgdb port
 */
void kgdb_putc(c)
	int	c;
{
	dcacnputc(1, c);
	delay(1000);
}
#endif

/*
 *	Execute a break instruction that will invoke kgdb
 */
void kgdb_break()
{
	if (!kgdb_enabled)
		return;

	asm(".export kgdb_break_inst,entry ! .label kgdb_break_inst");
	asm("break 0,5");
}

/*
 *	Enable a page for access, faulting it in if necessary
 */
boolean_t kgdb_mem_access(offset, access)
	vm_offset_t	offset;
	vm_prot_t	access;
{
	kern_return_t	code;

	/*
	 *	kernel_equiv_end is the end of equivalently mapped kernel
	 *	memory. if below then it's always accessible
	 */
	if (offset < kernel_equiv_end)
		return TRUE;
#if 1
	/* XXX ugly hack for negative numbers masquerading as addresses... */
	if ((offset & 0xFFF80000) == 0xFFF80000)
		return 0;

	return((pmap_extract(kernel_pmap, offset) == 0) ? 0 : 1);
#else
	/*
	 *    	fault in the memory just to make sure we can access it
	 */
	code = vm_fault(kernel_map, trunc_page(offset), access, FALSE, 
			FALSE, (void (*)()) 0);

	return code == KERN_SUCCESS;
#endif
}

/*
 *	See if we modified the kernel text and if so flush the caches.
 *	This routine is never called with a range that crosses a page
 *	boundary.
 */
void kgdb_flush_cache(offset, length)
	vm_offset_t	offset;
	vm_size_t	length;
{
#ifndef KGDB_LAN
	if (offset >= kernel_text_end)
		return;
#endif

#ifdef	CMU_PMAP
	offset = parisc_atop(offset);

	flush_dcache_region(offset, offset);
	purge_icache_region(offset, offset);
#else	/* CMU_PMAP */
	fdcache(0, offset, PARISC_PGBYTES);
	ficache(0, offset, PARISC_PGBYTES);
#endif	/* CMU_PMAP */
}


/*
 *	setup the thread to continue from the current point or a specified 
 *	address.
 */
void kgdb_continue(type, ssp, addr)
	int		type;
	struct parisc_saved_state *ssp;
	vm_offset_t	addr;
{
	if (addr != (vm_offset_t) 0) {
		ssp->iioq_head = addr;
		ssp->iioq_tail = addr + 4;
		if (kgdb_debug)
			printf("new pc = %8.8x\n", addr);
	}

	/*
	 *	Turn off the taken branch trap
	 */
	ssp->ipsw &= ~PSW_T;
}

/*
 *	setup the thread to single step from the current location or a
 *	specified address
 */
void kgdb_single_step(type, ssp, addr)
	int		type;
	struct parisc_saved_state *ssp;
	vm_offset_t	addr;
{
	Act		*inc = current_act();

	if (addr != (vm_offset_t) 0) {
		ssp->iioq_head = addr;
		ssp->iioq_tail = addr + 4;
		if (kgdb_debug)
			printf("new pc = %8.8x\n", addr);
	}

	/*
	 *	Save the old contents of the pc queue in the pcb so that we 
	 *	can restore them later.
	 */
	inc->mact.saved_oqt = ssp->iioq_tail;
	inc->mact.saved_sqt = ssp->iisq_tail;

	/*
	 *	Put a break instruction in the pc queue so that after executing
	 *	the queue head, we trap. 
	 */
	ssp->iioq_tail = (unsigned) user_sstep;
	ssp->iisq_tail = 0;

	/*
	 *	Set the taken branch trap bit in the PSW so that we can't 
	 *	branch around the breakpoint. Save the instruction at the head
	 *	of the queue in case we need to fix a link register after
	 *	a taken branch trap on a bl or blr.
	 */
	ssp->ipsw |= PSW_T;

	inc->mact.saved_insn = *(int *)(ssp->iioq_head & ~0x3);
	inc->mact.kgdb_trace |= KGDB_STEP;
}

/*
 *	fixup the pc if needed
 */
void kgdb_fixup_pc(type, ssp)
	int type;
	struct parisc_saved_state *ssp;
{
	unsigned	insn;
	Act		*inc;

	if (current_thread() &&
	    ((inc = current_act())->mact.kgdb_trace & KGDB_STEP)) {
		/*
		 *	If we performed a non-nullified branch-and-link,
		 *	the link register will point to user_nullify;
		 *	we must make it point after the original tail.
		 */
		if (type == I_TAKEN_BR) {
			insn = inc->mact.saved_insn;
			if ((insn & 0xfc000000) == 0xe4000000)	/* ble */
				ssp->r31 = inc->mact.saved_oqt + 4;
			else if ((insn & 0xfc000000) == 0xe8000000 &&
				 ((insn & 0xe000) == 0x4000 ||  /* blr */
				  (insn & 0xe000) == 0) &&      /* bl */
				 (insn & 0x3e00000) != 0) {	/*link != 0*/
				((int *)ssp)[(insn & 0x3e00000) >> 21] =
					inc->mact.saved_oqt + 4;
			}
			/* XXX also fiddle tail space on BE? */
		}

		/*
		 *	If we got to the break at user_nullify, we must have 
		 *	executed an instruction that nullified the user_sstep 
		 *	break but did not branch. In this case we must restore
		 *	the PSW N bit that was cleared at the successfully 
		 *	executed break.
		 */
		if (ssp->iioq_head == (unsigned) user_nullify)
			ssp->ipsw |= PSW_N;

		/*
		 *	If we didn't branch (the head or tail of the queue 
		 *	points after user_sstep), then the new tail follows 
		 *	the new head.
		 */
		if (ssp->iioq_head == (unsigned) user_nullify ||
		    ssp->iioq_tail == (unsigned) user_nullify) {
			ssp->iisq_tail = inc->mact.saved_sqt;
			ssp->iioq_tail = inc->mact.saved_oqt + 4;
		}

		/*
		 *	Restore the original succeeding pc.
		 */
		ssp->iisq_head = inc->mact.saved_sqt;
		ssp->iioq_head = inc->mact.saved_oqt;

		/*
		 *	Disable branch trap
		 */
		ssp->ipsw &= ~PSW_T;
		inc->mact.kgdb_trace &= ~KGDB_STEP;
	}

	/*
	 * If we arrived at kgdb_break, force us to continue
	 * past this point.
	 */
	if (ssp->iioq_head == (unsigned) kgdb_break_inst) {
		ssp->iioq_head += 4;
		if (ssp->iioq_head == ssp->iioq_tail)
			ssp->iioq_tail += 4;
	}
}

/*
 *	cause the kernel to dump core
 */
kgdb_dump_core()
{
	dumpsys();
}
#endif /* KGDB */
