/*
 * Copyright (c) 1988,1991,1994 University of Utah.
 * Copyright (c) 1982, 1986, 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by the University of California, Berkeley and its contributors''
 * in the documentation or other materials provided with the distribution
 * and in all advertising materials mentioning features or use of this
 * software. Neither the name of the University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: mem.c 1.6 94/11/10$
 */

/*
 * Memory special file
 */

#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#include <machine/machparam.h>
#include <machine/machspl.h>
#include <mach/vm_param.h>

mmopen(dev, flags)
	dev_t dev;
{
	return 0;
}

mmread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	return (mmrw(dev, uio, IO_READ));
}

mmwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	return (mmrw(dev, uio, IO_WRITE));
}

mmrw(dev, uio, rw)
	dev_t dev;
	struct uio *uio;
	int rw;
{
	int error = 0;
	int s, c;

	switch (minor(dev)) {

	case 0:			/* minor device 0 is physical memory */
		return (ENXIO);

	case 1:			/* minor device 1 is kernel memory */
		c = uio->io_count;
		
		if (!memaccess((caddr_t)uio->io_recnum, c, 0))
			return (EIO);

		if (rw == IO_WRITE) {
			/*
		 	 * Must be at splvm to keep out ISRs from trying to 
		 	 * look at the virtual address before we flush it. 
		 	 */
			
			s = splvm();
			bcopy(uio->io_data, (caddr_t)uio->io_recnum, c);
			fdcache(0, (caddr_t)uio->io_recnum, c);
			splx(s);
			return (error);
		}

		/*
		 * Allocate memory for read buffer.
		 */
		error = device_read_alloc(uio, c);
		if (error != KERN_SUCCESS)
			return (error);

		/*
		 * Must recheck access as we may have blocked in
		 * device_read_alloc.
		 * XXX do we need to undo the device_read_alloc?
		 */
		if (!memaccess((caddr_t)uio->io_recnum, c, 0))
			return (EIO);

		s = splvm();
		bcopy((caddr_t)uio->io_recnum, uio->io_data, c);
		fdcache(0, (caddr_t)uio->io_recnum, c);
		splx(s);
		return (error);
	}
}

/*
 * See if count bytes starting at VA addr are accessible.
 */
memaccess(addr, count, rw)
	unsigned int addr;
	int count, rw;
{
	extern pmap_t		kernel_pmap;
	extern unsigned int	equiv_end;
	register vm_offset_t	va;
	register vm_offset_t	vend = round_page(addr + count);

	if (addr >= 0 && vend < equiv_end)
		return 1;

	for (va = trunc_page(addr); va != vend; va += PAGE_SIZE)
		if (pmap_extract(kernel_pmap, va) == 0)
			return 0;
	return 1;
}
