/* 
 * Copyright (c) 1990-1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: rpb.h 1.6 94/12/14$
 */
#ifndef _RPB_H_
#define _RPB_H_

/*
 * The restart parameter block, which is a page in (very) low
 * core which runs after a crash.  Currently, the restart
 * procedure takes a dump.
 *
 * On the PA-RISC, the `rpb' looks like Processor Internal Memory.
 * `rp_sp' and `rp_flag' remain for HP-UX compatibility.
 */

#if !defined(LOCORE) && !defined(ASSEMBLER)
#include <machine/pim.h>

struct rpb {
	unsigned int rp_sp;	/* stack pointer */
	unsigned int rp_flag;	/* set when doadump runs (see below) */
	struct pim rp_pim;	/* Processor Internal Memory */
};

#define	rp_gr0	rp_pim.pim_gr[0]	/* start of general registers */
#define	rp_gr1	rp_pim.pim_gr[1]	/* gr0 is useless, start at gr1 */
#define	rp_sr0	rp_pim.pim_sr[0]	/* start of space registers */
#define	rp_cr0	rp_pim.pim_cr[0]	/* start of control registers */
#define	rp_cr8	rp_pim.pim_cr[8]	/* cr0-7 is useless, start at cr8 */

#ifdef KERNEL
extern	struct rpb rpb;
#endif
#endif	/* !(LOCORE||ASSEMBLER) */

/* rp_flags */
#define	RPB_DUMP	1	/* call to doadump() */
#define	RPB_B_DUMP	31		/* bit position for RPB_DUMP */
#define	RPB_CRASH	2	/* call to crashdump/doadump() */
#define	RPB_B_CRASH	30		/* bit position for RPB_CRASH */

#endif	/* _RPB_H_ */
