/* 
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: lan.c 1.6 94/11/10$
 *	Author: Bob Wheeler, Leigh Stoller, University of Utah CSL
 */

/*
 * 	HP 27125 CIO LAN driver
 */

#include <ciolan.h>
#define NLAN NCIOLAN

#if NLAN > 0
#include <device/device_types.h>
#include <device/io_req.h>
#include <device/if_hdr.h>
#include <device/if_ether.h>
#include <device/net_status.h>
#include <device/net_io.h>

#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#include <machine/machparam.h>
#include <machine/spl.h>

#include <hpcio/device.h>
#include <hpcio/cioreg.h>
#include <hpcio/lanreg.h>
#include <hpcio/lancons.h>

struct lan_softc {
	struct ifnet	    is_if;	/* generic interface header */
	unsigned char       addr[6];	/* hardware address */
	struct cio_chan	    *cc;	/* pointer to cio channel struct */
	struct dmaquad	    *cmdq;	/* generic command quad */
	struct dmaquad	    *cmdsq;	/* generic command RS quad */
	struct dmaquad	    *xmitq;	/* transmit quad */
	struct dmaquad	    *xmitsq;	/* transmit RS quad */
	struct dmaquad	    *recvq0;	/* receive quad */
	struct dmaquad	    *recvq1;	/* receive quad */
	struct dmaquad	    *recvsq;	/* receive RS quad */
	struct dmaquad	    *quads;	/* scatter/gather quads */
	struct linkquad     *reclink;	/* recieve link quad (for count) */
	struct statquad	    *recstat;	/* status quad for receive DMA chain */
	struct semaquad     *recsema;	/* semaphore quad for receive */
	struct linkquad     *link;	/* link quad */
	struct statquad	    *stat;	/* status quad for DMA chain */
	struct semaquad     *sema;	/* semaphore quad */
	struct cio_request  *cr;	/* cio request packet */
	int		    xmitbufs;	/* Number of free transmit buffers */
	int                 alive;	/* alive and configured */
};

struct lan_softc lan_softc[NLAN];
struct cio_request lancr[NLAN];

int	lanstart();
int	lanattach(), lanprobe(), lanscan(), lanlayer();
void	lanintr(), landisplay();
struct driver ciolandriver = {
	"lan", lanattach, lanprobe, lanscan, landisplay, lanlayer
};

lanprobe(cc)
	struct cio_chan *cc;
{
	return (cc->type == CIOIDY_LAN) ? 0 : ENXIO;
}

lanattach(cc, unit, flags)
	struct cio_chan *cc;
{
	register struct lan_softc *ls = &lan_softc[unit];
	struct dmaquad *q;
	struct lan_interface *li;
	struct ifnet *is_if = &ls->is_if;
	int i;
	u_char *bp;

	if (cc->alive)
		return 0;

	if (unit > NLAN)
		return ENXIO;

	/*
	 * remember the channel that we are assigned on the CIO
	 */
	ls->cc = cc;

	/* 
	 * Allocate buffers and DMA quads. We will allocate a set of quads
	 * for both transmit and receive operations.
	 */
	q = (struct dmaquad *) cio_malloc(sizeof(struct dmaquad) * 15,
					 CIO_QUAD_ALIGN);
	ls->cmdq    = q++;
	ls->cmdsq   = q++;
	ls->xmitq   = q++;
	ls->xmitsq  = q++;
	ls->recvq0  = q++;
	ls->recvq1  = q++;
	ls->recvsq  = q++;
	ls->reclink = (struct linkquad *) q++;
	ls->recstat = (struct statquad *) q++;
	ls->recsema = (struct semaquad *) q++;
	ls->link    = (struct linkquad *) q++;
	ls->stat    = (struct statquad *) q++;
	ls->sema    = (struct semaquad *) q++;
	ls->quads   = q;

	/*
	 * this link quad is used by all operations
	 */
	ls->link->link      = CIO_EOC(ls->link);
	ls->link->command   = CIODMA_LNKST;
	ls->link->sema      = ls->sema;
	ls->link->stat      = ls->stat;

	/*
	 * general quads for initialization
	 */
	ls->cmdq->link      = ls->cmdsq;
	ls->cmdq->command   = CIODMA_WC | CIO_BYTE;
	ls->cmdq->address   = (caddr_t) cio_malloc(CIO_MIN_BUFFER, 
						CIO_BUFFER_ALIGN);

	ls->cmdsq->link     = (struct dmaquad *) ls->link;
	ls->cmdsq->command  = CIODMA_RS | CIO_BYTE | CIOLAN_RS_INTERFACE;
	ls->cmdsq->count    = CIO_MIN_BUFFER * 5;
	ls->cmdsq->address  = (caddr_t) cio_malloc(CIO_MIN_BUFFER, 
						   CIO_BUFFER_ALIGN);

	/*
	 * we cannot guarantee the alignment of the data packets so we will 
	 * keep two large buffers around, one to send from and one to receive
	 * from.
	 */
	ls->xmitq->link     = ls->xmitsq;
	ls->xmitq->command  = CIODMA_WD | CIO_BYTE;
	ls->xmitq->address  = (caddr_t) cio_malloc(CIOLAN_BUFFER_SIZE,
						 CIO_BUFFER_ALIGN);

	ls->xmitsq->link    = (struct dmaquad *) ls->link;
	ls->xmitsq->command = CIODMA_RS | CIO_BYTE | CIOLAN_RS_INTERFACE;
	ls->xmitsq->count   = CIO_MIN_BUFFER;
	ls->xmitsq->address = (caddr_t) cio_malloc(CIO_MIN_BUFFER, 
						CIO_BUFFER_ALIGN);
	/*
	 * It has been empirically shown that if you want to request a full
	 * ethernet packet with status that you have to do it with two quads.
	 */
	ls->recvq0->link    = ls->recvq1;
	ls->recvq0->command = CIODMA_RD | CIO_BYTE | CIOLAN_RD_ARM | CIO_CONT;
	ls->recvq0->address = (caddr_t) cio_malloc(CIOLAN_BUFFER_SIZE,
						 CIO_BUFFER_ALIGN);
	ls->recvq0->count   = 0x80;

	ls->recvq1->link    = (struct dmaquad *) ls->reclink;
	ls->recvq1->command = CIODMA_RD | CIO_BYTE | CIOLAN_RD_ARM;
	ls->recvq1->address = ls->recvq0->address + 0x80;
	ls->recvq1->count   = CIOLAN_BUFFER_SIZE - 0x80;

	/*
	 * we need a link quad here to get the size of the packet received
	 */
	ls->reclink->link    = ls->recvsq;
	ls->reclink->command = CIODMA_LNKST;
	ls->reclink->sema    = ls->recsema;
	ls->reclink->stat    = ls->recstat;

	ls->recvsq->link     = (struct dmaquad *) ls->link;
	ls->recvsq->command  = CIODMA_RS | CIO_BYTE | CIOLAN_RS_INTERFACE;
	ls->recvsq->count    = CIO_MIN_BUFFER;
	ls->recvsq->address  = (caddr_t) cio_malloc(CIO_MIN_BUFFER, 
						  CIO_BUFFER_ALIGN);

	/*
	 * now fill in the CIO request packet
	 */
	ls->cr = &lancr[unit];
	ls->cr->stat    = ls->stat;
	ls->cr->quad    = ls->quads;
	ls->cr->ciounit = cc->ciounit;
	ls->cr->channel = cc->channel;
	ls->cr->polled  = 0;

	cc->debug = 0;
	cc->alive = 1;
	ls->alive = 1;

	/*
	 * install the interrupt service routine
	 */
	cc->int_level = SPL_IBIT_CIOLAN;
	(void)procitab(SPLCIOLAN, lanintr, 0, 0);

	/*
	 * LAN Initialization
	 */
	ls->cr->cmdq = q = ls->cmdq;

	/*
	 * cause LAN to go offline if it isn't already
	 */
	bp = (u_char *) q->address;
	lan_go_offline(bp);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 0)) {
		printf("lanattach: lan%d go offline initialize error\n", unit);
		cc->alive = 0;
		ls->alive = 0;
		return EIO;
	}

	/*
	 * read the node address from the last read sense packet
	 */
	li = (struct lan_interface *) ls->cmdsq->address;
	for (i = 0; i < 6; i++)
		ls->addr[i] = li->hw_addr[i];

	/*
	 * Perform external loopback test
	 */
	bp = (u_char *) q->address;
	lan_loopback_test(bp);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 0)) {
		printf("lanattach: lan%d loopback test error\n", unit);
		cc->alive = 0;
		ls->alive = 0;
		return EIO;
	}

	/*
	 * fill in interface structure
	 */
	is_if->if_address       = (char *) &ls->addr;
	is_if->if_unit          = unit;
	is_if->if_header_size   = sizeof(struct ether_header);
	is_if->if_header_format = HDR_ETHERNET;
	is_if->if_address_size  = 6;
	is_if->if_mtu           = ETHERMTU;
	is_if->if_flags         = IFF_BROADCAST;

	if_init_queues(is_if);

	return 0;
}

lanscan(unit)
{
	return 0;
}

void
landisplay(unit)
{
	register struct lan_softc *ls = &lan_softc[unit];

	iprintf("lan%d hardware address : %x:%x:%x:%x:%x:%x\n", unit,
		ls->addr[0], ls->addr[1], ls->addr[2], ls->addr[3], 
		ls->addr[4], ls->addr[5]);
}
	       
lanlayer(unit, layer, devname)
	int *layer;
	char *devname;
{
	register struct lan_softc *ls = &lan_softc[unit];

	if (*layer >= NLAN)
		return ENXIO;

	if (ls->alive) {
		*devname++ = 'l';
		*devname++ = 'a';
		*devname++ = 'n';
		*devname++ = unit % 10 + '0';
		return 0;
	}

	return ENXIO;
}

laninit(unit)
{
	register struct lan_softc *ls = &lan_softc[unit];
	struct ifnet *is_if = &ls->is_if;
	struct dmaquad *q;
	int s;
	u_char *bp;

	if (unit > NLAN)
		return ENXIO;

	/*
	 * set up cio request for command
	 */
	ls->cr->cmdq = q = ls->cmdq;

	/*
	 * Config filters (multicast and broadcast).
	 */
	bp = (u_char *) q->address;
	lan_set_filter(bp, CIOLAN_FLT_B|CIOLAN_FLT_M);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 0)) {
		printf("laninit: lan%d set packet filters error\n", unit);
		return EIO;
	}

	/*
	 * Set ARQ limits (ie. allow interrupt driven packet input).
	 */
	bp = (u_char *) q->address;
	lan_arm_receive(bp, 1, 1);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 0)) {
		printf("laninit: lan%d set ARQ limits error\n", unit);
		return EIO;
	}

	/*
	 * cause LAN to go online
	 */
	bp = (u_char *) q->address;
	lan_go_online(bp);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 1)) {
		printf("laninit: lan%d go online initialize error\n", unit);
		return EIO;
	}

	s = splimp();
	is_if->if_flags |= IFF_RUNNING | IFF_UP;

	/*
	 * Arm the packet receive interrupt by looking for the first packet.
	 */
	ls->cr->cmdq = ls->recvq0;
	ls->cr->polled = 0;
	if (ciosync(ls->cr) || lan_error(ls->recvsq->address, unit, 1)) {
		printf("laninit: lan%d get ready for packets error\n", unit);
		return EIO;
	}
	splx(s);

	ls->xmitbufs = 8;
	
	return 0;
}

lanopen(dev, flag)
{
	int unit = minor(dev);
	struct lan_softc *ls = &lan_softc[unit];

	if (unit > NLAN || !ls->alive)
		return ENXIO;

	if ((ls->is_if.if_flags & IFF_UP) == 0) {
		int status = laninit(unit);
		if (status)
			return status;
		ls->is_if.if_flags |= IFF_UP;
	}
	return 0;
}

lanwrite(dev, ior)
	dev_t	dev;
	io_req_t ior;
{
	int unit = minor(dev);
	struct lan_softc *ls = &lan_softc[unit];
	int status;

	if (unit > NLAN || !ls->alive)
		return ENXIO;

	return net_write(&ls->is_if, lanstart, ior);
}

lansetinput(dev, receive_port, priority, filter, filter_count)
	dev_t		dev;
	mach_port_t	receive_port;
	int		priority;
	filter_t	filter[];
	unsigned int	filter_count;
{
	register int unit = minor(dev);
	struct lan_softc *ls = &lan_softc[unit];

	if (unit > NLAN || !ls->alive)
		return ENXIO;

	return net_set_filter(&ls->is_if, receive_port, priority,
			filter, filter_count);
}

langetstat(dev, flavor, status, count)
	dev_t	dev;
	int	flavor;
	dev_status_t	status;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	register int unit = minor(dev);
	struct lan_softc *ls = &lan_softc[unit];
	
	if (unit > NLAN || !ls->alive)
		return ENXIO;

	return net_getstat(&ls->is_if, flavor, status, count);
}

lansetstat(dev, flavor, status, count)
	dev_t	dev;
	int	flavor;
	dev_status_t	status;
	unsigned int	count;
{
	register int unit = minor(dev);
	struct lan_softc *ls = &lan_softc[unit];
	
	if (unit > NLAN || !ls->alive)
		return ENXIO;

	switch (flavor) {
	case NET_STATUS: {
		/*
		 * All we can change are flags, and not many of those.
		 */

		/*
		 * Set interface flags.  Possibilities are: bringing
		 * the interface up or down and setting it into or out
		 * of promiscuous mode.
		 */
		register struct net_status *ns = (struct net_status *)status;
		register int	up;
		register int	promisc;
		short old_flags = ls->is_if.if_flags;

		if (count < NET_STATUS_COUNT)
			return D_INVALID_OPERATION;

		up = ns->flags & IFF_UP;
		promisc = ns->flags & IFF_PROMISC;
		if (  up != (old_flags & IFF_UP) 
		    || promisc != (old_flags & IFF_PROMISC)) {
			struct ifnet *ifp = &ls->is_if;

			ifp->if_flags = ns->flags;
			if (!up)
				ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
			else
				laninit(unit);

			if (ifp->if_flags & IFF_UP)
				ifp->if_flags |= IFF_RUNNING;
			else
				ifp->if_flags &= ~(IFF_RUNNING | IFF_PROMISC);

			if (promisc & ! (ifp->if_flags & IFF_PROMISC))
				return D_DEVICE_DOWN;
		}
		break;
	}
	default:
		return (D_INVALID_OPERATION);
	}
	return D_SUCCESS;
}

lanreset(unit)
{
	register struct lan_softc *ls = &lan_softc[unit];
	struct dmaquad *q;
	u_char *bp;

	if (unit > NLAN)
		return ENXIO;

	/*
	 * set up cio request for command
	 */
	ls->cr->cmdq = q = ls->cmdq;

	/*
	 * cause LAN to go offline if it is not already
	 */
	bp = (u_char *) q->address;
	lan_go_offline(bp);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 0)) {
		printf("lanreset: lan%d go offline initialize error\n", unit);
		return EIO;
	}

	/*
	 * soft reset LAN
	 */
	bp = (u_char *) q->address;
	lan_soft_reset(bp);
	q->count = bp - (u_char *) q->address;

	if (ciosync(ls->cr) || lan_error(ls->cmdsq->address, unit, 0)) {
		printf("lanreset: lan%d soft reset initialize error\n", unit);
		return EIO;
	}

	ls->is_if.if_flags &= ~IFF_RUNNING;

	return laninit(unit);
}

lanstart(unit)
{
	struct lan_softc *ls = &lan_softc[unit];
	io_req_t m;
	struct ether_header *header;
	int status;

	IF_DEQUEUE(&ls->is_if.if_snd, m);

	while (m != (io_req_t) 0) {
		if (ls->xmitbufs == 0) {
			printf("lanstart: lan%d No free transmit buffers\n",
			       unit);
			m->b_flags |= B_ERROR;
			m->b_error = EIO;
			m->b_resid = m->b_bcount;
			iodone(m);
			return 0;
		}

		if (m->io_count > CIOLAN_MAX_TU)
			status = EIO;
		else {
			bcopy(m->io_data, ls->xmitq->address, m->io_count);

			if (m->io_count < CIOLAN_MIN_TU)
				m->io_count = CIOLAN_MIN_TU;

			ls->xmitq->count = m->io_count;
			status = lantransmit(unit);
		}

		if (status) {
			/*
			 * there was an error queuing the last request, get 
			 * rid of it and then look for another request
			 */
			m->b_flags |= B_ERROR;
			m->b_error = EIO;
			m->b_resid = m->b_bcount;
			iodone(m);
			return 0;
		}
		iodone(m);
		/*
		 * loop until the queue is empty.
		 */
		IF_DEQUEUE(&ls->is_if.if_snd, m);
	}
	/*
	 * no more packets to send
	 */
	return 0;
}

lantransmit(unit)
{
	struct lan_softc *ls = &lan_softc[unit];
	struct lan_interface *li = (struct lan_interface *)ls->xmitsq->address;
	int s;

	s = splimp();
	ls->cr->cmdq = ls->xmitq;

	if (ciosync(ls->cr) || lan_error(ls->xmitsq->address, unit, 1)) {
		splx(s);
		return EIO;
	}
	splx(s);

	/*
	 * Set transmit buffer count.
	 */
	ls->xmitbufs = li->xmit_buffers;
	
	return 0;
}

void
lanintr()
{
	struct lan_softc *ls;
	struct lan_header *lh;
	struct ether_header *header;
	int unit;
	ipc_kmsg_t kmsg;
	int length, s;
	char *start;

	for (unit = 0; unit < NLAN; unit++) {
		struct lan_softc *ls = &lan_softc[unit];

		if (!ls->alive)
			continue;

		/*
		 * see if there is an interrupt on the CIO card for this lan
		 */
		if (test_cio_are(ls->cc) == 0)
			continue;

		lh = (struct lan_header *) ls->recvq0->address;
		header = &lh->hdr;

		do {
			ls->cr->cmdq = ls->recvq0;

			if (ciosync(ls->cr) || 
			    lan_error(ls->recvsq->address, unit, 1))
				break;

			/*
			 * if this is an empty packet then don't process it...
			 */
			if (ls->recvq0->count + ls->recvq1->count 
			    - ls->recstat->count <= 6)
				continue;

			/*
			 * Set transmit buffer count for output routine.
			 */
			ls->xmitbufs = lh->xmit_buffers;

			/*
			 * this is a valid packet. Get a new kmsg to pass it in
			 */
			kmsg = net_kmsg_get();
			
			/*
			 * If no kmsg then drop the received packet
			 */
			if (kmsg == IKM_NULL)
				continue;

			length = ls->recvq0->count + ls->recvq1->count - 
				ls->recstat->count - 6 -
					sizeof(struct ether_header);
			start = (char *)(net_kmsg(kmsg)->packet +
					 sizeof(struct packet_header));
		
			bcopy(lh->data, start, length);
			bcopy(header, net_kmsg(kmsg)->header,
			      sizeof(struct ether_header));

			length += sizeof(struct packet_header);
		
			((struct packet_header *)
			 &net_kmsg(kmsg)->packet[0])->length = length;
			((struct packet_header *)
			 &net_kmsg(kmsg)->packet[0])->type = 
				 header->ether_type;

			net_packet(&ls->is_if, kmsg, length, 
				   ethernet_priority(kmsg));

		} while (lh->rec_buffers > 1);

		clear_cio_are(ls->cc);
	}

}

static char *err2[] = {
	"LANCE cannot access receive buffer",
	"CRC error",
	"overflow on receive",
	"framing error",
	"retry failure on transmit",
	"lost carrier",
	"late collision detected",
	"underflowed",
	"LANCE cannot access transmit packet",
	"LANCE cannot access shared data structures",
	"No transciever heartbeat sensed",
	"LANCE detected babble error",
	"external loopback test failed",
	"both NVRAM banks have failed",
	"10 ms timer selftest failed",
	"LANCE internal selftest failed"
};

lan_error(s, unit, check_online)
	struct lan_interface *s;
{
	int status = 0;
	int i;

	/*
	 * we can optionally check for the lan being offline
	 */
	if (check_online && s->status & CIOLAN_IE_NOL) {
		printf("lan%d is not online\n", unit);
		status = EIO;
	}

	if (s->status & 0x8300) {
		status = EIO;
			
		if (s->status & CIOLAN_IE_STF)
			printf("lan%d selftest failed\n", unit);

		if (s->status & CIOLAN_IE_PFB)
			printf("lan%d power fuse blown\n", unit);

		if (s->status & CIOLAN_IE_NTH)
			printf("lan%d no 802.3 transceiver heartbeat\n", unit);
	}

	if (s->selftest) {
		status = EIO;
		for (i = 0; i < 16; i++)
			if (s->status & (1 << i))
				printf("lan%d error: %s\n", unit, err2[i]);
	}

	return status;
}
#endif NLAN > 0
