/* 
 * Copyright (c) 1991,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: muxreg.h 1.2 91/09/25$
 * 	Author: Leigh Stoller, Bob Wheeler, University of Utah CSL
 */

/*
 *	Macros and definitions for CIO 6 port serial mux card
 */

#define MUX_OUT_BUFSIZE	8192	/* size of I/O buffer for output */
#define MUX_IN_BUFSIZE	8192	/* size of I/O buffer for input  */

/*
 * macros to build mux commands in output buffer
 */
#define mux_start_data_block(bufp, port, len) 		\
	do {						\
		*(bp)++ = (port);   			\
	        *(bp)++ = 0;   				\
	        *(bp)++ = len;				\
	} while(0)

#define mux_end_block(bp)				\
	do {						\
		*(bp)++ = 0xff;				\
	} while(0)

#define mux_transmittion_mode(bp, port, mask)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 1;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (mask);		\
	} while(0)

#define AUTO_ANSWER		0x00
#define ORIGINATE		0x10
#define FULL_DUPLEX_MODEM	0x01
#define FULL_DUPLEX_HARDWIRED	0x02

#define mux_baudrate(bp, port, rate)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 2;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (rate);		\
	} while(0)

#define	BAUD_50		0x01
#define	BAUD_75		0x02
#define	BAUD_110	0x03
#define	BAUD_134	0x04
#define	BAUD_150	0x05
#define	BAUD_300	0x06
#define	BAUD_600	0x07
#define	BAUD_900	0x08
#define	BAUD_1200	0x09
#define	BAUD_1800	0x0a
#define	BAUD_2400	0x0b
#define	BAUD_3600	0x0c
#define	BAUD_4800	0x0d
#define	BAUD_7200	0x0e
#define	BAUD_9600	0x0f
#define	BAUD_19200	0x10

#define mux_char_length(bp, port, mask)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 3;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (mask);		\
	} while(0)

#define BITS_5		0x00
#define BITS_6		0x10
#define BITS_7		0x20
#define BITS_8		0x30
#define STOP_1		0x00
#define STOP_1_5	0x04
#define STOP_2		0x08
#define PARITY_NONE	0x00
#define PARITY_ODD	0x01
#define PARITY_EVEN	0x02


#define mux_modem_connect_timer(bp, port, seconds)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 4;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (seconds);		\
	} while(0)

#define mux_modem_no_activity_timer(bp, port, seconds)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 5;				\
		*(bp)++ = 2;				\
		*(bp)++ = (u_char) (seconds >> 8 & 0xff);\
		*(bp)++ = (u_char) (seconds & 0xff);	\
	} while(0)

#define mux_lost_receiver_timer(bp, port, ten_msec)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 6;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (ten_msec);		\
	} while(0)

#define mux_hardware_xon_xoff(bp, port, xon, xoff)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 7;				\
		*(bp)++ = 2;				\
		*(bp)++ = (u_char) (xon);		\
		*(bp)++ = (u_char) (xoff);		\
	} while(0)

#define	XON_DISABLE	0xff
#define	XOFF_DISABLE	0xff

#define mux_host_xon_xoff(bp, port, xon, xoff)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 8;				\
		*(bp)++ = 2;				\
		*(bp)++ = (u_char) (xon);		\
		*(bp)++ = (u_char) (xoff);		\
	} while(0)

#define mux_open_modem_link(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 128;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_close_modem_link(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 129;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_start_break(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 130;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_stop_break(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 131;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_clear_rts_line(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 132;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_set_rts_line(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 133;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_clear_dtr_line(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 134;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_set_dtr_line(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 135;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_clear_signal_rate_line(bp, port)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 136;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_set_signal_rate_line(bp, port)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 137;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_flush_xmit_buffer(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 138;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_stop_xmit(bp, port)				\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 139;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_restart_xmit(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 140;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_get_modem_status(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 141;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_get_connect_status(bp, port)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 142;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_get_ram_contents(bp, port, address, length)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 143;				\
		*(bp)++ = 3;				\
		*(bp)++ = (u_char) (address >> 8 & 0xff);\
		*(bp)++ = (u_char) (address & 0xff);	\
		*(bp)++ = (u_char) (length);		\
	} while(0)

#define mux_set_host_int_mask(bp, port, mask)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 146;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (mask);		\
	} while(0)

#define	MODEM_CHANGES	0x08
#define	DATA_PRESENT	0x01

#define mux_disable_receiver(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 147;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_enable_receiver(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 148;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_send_host_xon(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 149;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_send_host_xoff(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 150;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_disable_modem_changes_quads(bp, port)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 151;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_enable_modem_changes_quads(bp, port)	\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 152;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_set_modem_int_mask(bp, port, mask)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 153;				\
		*(bp)++ = 1;				\
		*(bp)++ = (u_char) (mask);		\
	} while(0)

#define CLEAR_TO_SEND		0x80
#define RING_INDICATOR		0x40
#define DATA_CARRIER_DETECT	0x20
#define DATA_SET_READY		0x10

#define mux_set_count_always(bp, port)			\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 154;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_set_count_on_change(bp, port)		\
	do {						\
		*(bp)++ = (u_char) (port) | 0x80;	\
		*(bp)++ = 155;				\
		*(bp)++ = 0;				\
	} while(0)

/*
 * macros for down loading card
 */

#define mux_initialize_download(bp)			\
	do {						\
		*(bp)++ = 1;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_set_download_address(bp, add, len)		\
	do {						\
		*(bp)++ = 2;				\
		*(bp)++ = 8;				\
		*(bp)++ = (u_char) (add >> 24 & 0xff);	\
		*(bp)++ = (u_char) (add >> 16 & 0xff);	\
		*(bp)++ = (u_char) (add >> 8  & 0xff);	\
		*(bp)++ = (u_char) (add & 0xff);	\
		*(bp)++ = (u_char) (len >> 24 & 0xff);	\
		*(bp)++ = (u_char) (len >> 16 & 0xff);	\
		*(bp)++ = (u_char) (len >> 8  & 0xff);	\
		*(bp)++ = (u_char) (len & 0xff);	\
	} while(0)

#define mux_verify_download(bp)				\
	do {						\
		*(bp)++ = 3;				\
		*(bp)++ = 0;				\
	} while(0)

#define mux_start_execution(bp, add)			\
	do {						\
		*(bp)++ = 4;				\
		*(bp)++ = 4;				\
		*(bp)++ = (u_char) (add >> 24 & 0xff);	\
		*(bp)++ = (u_char) (add >> 16 & 0xff);	\
		*(bp)++ = (u_char) (add >> 8  & 0xff);	\
		*(bp)++ = (u_char) (add & 0xff);	\
	} while(0)


/*
 * Return codes in input buffer
 */
#define	MUX_PORT_COUNT	128
#define MUX_MODEM_STATE 141

/*
 * bits in status byte for each charater
 */
#define MUX_STATUS_OVERFLOW	0x20
#define MUX_STATUS_MODEM	0x10
#define MUX_STATUS_BREAK	0x08
#define MUX_STATUS_FRAME	0x04
#define MUX_STATUS_OVERRUN	0x02
#define MUX_STATUS_PARITY	0x01

/*
 * modem state bits for modem status 141
 */
#define	MUX_CTS	0x80
#define	MUX_RI	0x40
#define	MUX_DCD	0x20
#define	MUX_DSR	0x10
#define	MUX_RTS	0x04
#define	MUX_DTR	0x81

/*
 * micellaneous masks
 */
#define MUX_ENDBLOCK	0xff
#define MUX_STATUS_MASK	0x80
#define MUX_PORT_MASK	0x7f

/* 
 * Size of transmit buffer on card. Docs say something about leaving 2 open 
 */
#define TBUFSIZE	(256-2)

/*
 * codes for muxctl
 */
#define	MUX_ENABLE	0
#define MUX_BREAK	1
