/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * unasm.c -- HP_PA Instruction Printer
 */

#include <machine/kdbdefs.h>

/*
 *  $Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel_unused/parisc/kdb/unasm.c,v 1.5 1994/07/21 22:32:05 mike Exp $
 *
 *  Spectrum Simulator Datatype Definitions
 *  Dan Magenheimer - 4/5/82
 *  Computer Research Center, Hewlett-Packard Labs
 */

typedef	int		VOID, BOOL, WORD;
typedef	char		*TEXT;

#define		FAST		register

#define		EXPORT
#define		IMPORT		extern

#undef		NULL
#define		NULL		0
#define		NO		0
#define		YES		1

/* IMPORTANT NOTE:
 *  All modules using this header may assume that the datatype "int" is a
 *   32-bit (or > 32-bit) signed quantity.
 */


/* Spectrum Architecturally Defined Datatypes */
struct doubleword {
	WORD	wd0;
	WORD	wd1;
};

struct quadword {
	struct	doubleword	d0;
	struct	doubleword	d1;
};

	/* datatypes for halfword and byte fields of a word are defined
	 *  in ssBits.h */

/* Memory addressing datatypes */
typedef	unsigned int	SID,	/* range [0..MAXSID] */
			PGID,	/* range [0..MAXPGID] */
			OFS,	/* range [0..MAXINT]  */
			REALADR; /* range [0..MAXINT] */


/* data sizes */
enum datasize { Byte, Halfword, Word, Doubleword, Quadword, Variable };

/* Miscellaneous datatypes */
typedef	unsigned int	FLAGS;

/* struct for entry in unwind table */
struct ute {
	WORD	word1;
	WORD	word2;
	WORD	word3;
	WORD	word4;
};
/*
 *  $Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel_unused/parisc/kdb/unasm.c,v 1.5 1994/07/21 22:32:05 mike Exp $
 *
 *  Spectrum Instruction Set Condition Completer Bit Assignments
 *  Dan Magenheimer - 6/14/82
 *  Terrence Miller - 6/21/82
 *  Computer Research Center, Hewlett-Packard Labs
 *
 *  (c) copyright 1982
 *  (p) protected 1982
 *  The Hewlett-Packard Company
 *  Hewlett-Packard Laboratories
 *  Computer Research Center
 *  Palo Alto, California
 *
 *  *** HP Company Confidential ***
 *
 * $Log: unasm.c,v $
 * Revision 1.5  1994/07/21  22:32:05  mike
 * official HP copyright notice
 *
 * Revision 1.4  1992/07/08  12:19:52  dalton
 * Checkin before split to 1.0.4 release (by LBS).
 *
 * Revision 1.3  92/06/06  16:16:45  dalton
 * *** empty log message ***
 * 
 * Revision 1.2  92/06/06  15:42:28  dalton
 * Changed include to be a path relative to hp800.
 * 
 * Revision 1.1  92/06/06  14:05:33  dalton
 * Initial revision
 * 
 * Revision 1.2  91/04/14  20:29:49  osfrcs
 * 	Initial version.
 * 	[91/03/30  09:20:34  brezak]
 * 
 * Revision 1.1.2.2  91/04/02  10:42:50  brezak
 * 	Initial version.
 * 	[91/03/30  09:20:34  brezak]
 * 
 * Revision 1.1.1.2  91/03/30  09:20:34  brezak
 * 	Initial version.
 * 
;Revision 1.1  88/07/11  14:05:15  14:05:15  ren (Bob Naas)
;Initial revision
;
 * Revision 5.2  87/07/02  14:45:57  14:45:57  kent (Kent McMullen)
 * added constants to support addDasm and addDCond added to ssDID.c
 * 
 * Revision 5.1  87/02/27  11:12:08  11:12:08  kent (Kent McMullen)
 * update all src to 5.1
 * 
 * Revision 5.0  87/02/18  16:31:15  16:31:15  kent (Kent McMullen)
 * update all revision numbers to 5.0 for release
 * 
 * Revision 1.1  86/07/15  08:34:55  08:34:55  kent (Kent McMullen)
 * Initial revision
 * 
 * Revision 4.1  83/10/25  17:01:22  17:01:22  djm (Daniel J Magenheimer)
 * First release for ACD v4
 * 
 * Revision 3.0  83/06/13  10:22:59  djm (Daniel Magenheimer)
 * First release for distribution
 * 
 *
 */


/* Arithmetic/Logical Conditions */
#define	NEV	0x0
#define	EQZ	0x2
#define	LT	0x4
#define	LE	0x6
#define	LLT	0x8
#define	NUV	0x8
#define	LLE	0xA
#define	ZNV	0xA
#define	SV	0xC
#define	OD	0xE
#define	TR	0x1
#define	NEQZ	0x3
#define	GE	0x5
#define	GT	0x7
#define	LGE	0x9
#define	UV	0x9
#define	LGT	0xB
#define	VNZ	0xB
#define	NSV	0xD
#define	EV	0xF

/* unit conditions */
#define	SBZ	0x4
#define	SHZ	0x6
#define	SDC	0x8
#define	SBC	0xC
#define	SHC	0xE
#define	NBZ	0x5
#define	NHZ	0x7
#define	NDC	0x9
#define	NBC	0xD
#define	NHC	0xF

/*field conditions */
#define XEQ	0x1
#define XLT	0x2
#define	XOD	0x3
#define XTR	0x4
#define XNE	0x5
#define XGE	0x6
#define XEV	0x7



/*
 *  These macros are designed to be portable to all machines that have
 *  a wordsize greater than or equal to 32 bits that support the portable
 *  C compiler and the standard C preprocessor.  Wordsize (default 32)
 *  and bitfield assignment (default left-to-right,  unlike VAX, PDP-11)
 *  should be predefined using the constants HOSTWDSZ and BITFRL and
 *  the C compiler "-D" flag (e.g., -DHOSTWDSZ=36 -DBITFLR for the DEC-20).
 *  Note that the macro arguments assume that the integer being referenced
 *  is a 32-bit integer (right-justified on the 20) and that bit 0 is the
 *  most significant bit.
 */

#ifndef HOSTWDSZ
#define	HOSTWDSZ	32
#endif

#ifdef	vax
#ifndef BITFLR
#define	BITFRL
#endif
#else
#define	BITFLR
#endif


/*###########################  Macros  ######################################*/

/*---------------------------------------------------------------------------
 * DeclareBitfield$Reference - Declare a structure to be used to reference
 *  a specified bitfield within an integer (using BitfR, see below).
 *  The argument "n" must be an identifier name not used elsewhere in the
 *  program , "s" and "l" must (alas!) be constants.  (Suggestion: if
 *  "s" == 2 and "l" == 8, use _b28 for "n".)  The name "BITFLR" should
 *  be pre-defined if the compiler assigns bitfields from left-to-right.
 *  The resultant macro expansion defines a stucture in which the bit field
 *  starting at position "s" with length "l" may be referenced by accessing
 *  member "n".  [Note: The leftmost bits in a 36-bit word may be accessed
 *  by specifying -4 <= s < 0 on the DEC-20.]
 *---------------------------------------------------------------------------*/

#ifdef	BITFRL
#define	DeclBitfR(s,l,n) struct n { int:(HOSTWDSZ-(s)-(l)); unsigned n:l;};
#else
#define	DeclBitfR(s,l,n) struct n { int:((s)+(HOSTWDSZ-32)); unsigned n:l;};
#endif

/*---------------------------------------------------------------------------
 * Bitfield$Reference - Reference a specified bitfield within an integer.
 *  The argument "i" must be an addressable variable (i.e., not a register
 *  variable or an expression... but see BitfX below), "n" must be an
 *  identifier name declared in a DeclBitfR invocation.  The resultant
 *  macro expansion references the bit field in "i" described by the
 *  DeclBitfR invocation with the same name ("n").  BitfR may be used as
 *  an lvalue or an rvalue. (i.e., either side of an assignment statement)
 *  The "s" and "l" arguments are historical and are now unused.  (They
 *  correspond to the "s" and "l" arguments in DeclBitfR)
 *  Translates to a single instruction on both the VAX and the DEC-20.
 *---------------------------------------------------------------------------*/
#define	BitfR(i,s,l,n)	(((struct n *)&i)->n)

/*---------------------------------------------------------------------------
 * Bitfield$eXtract - Extract the specified field from an integer.  Arguments
 *  are the same as for BitfR (except no "n"), however both "s" and "l" need
 *  no longer be constants. May only be used as an rvalue. Translates to
 *  two instructions on the VAX, three on the DEC-20.
 *---------------------------------------------------------------------------*/

#define	BitfX(i,s,l)	 (((i) >> (32-(s)-(l))) & ((1 << (l)) - 1))


/*---------------------------------------------------------------------------
 * Mask$32bits - Mask the low order 32 bits of passed word.  No-op on 32
 *  bit machines.
 *---------------------------------------------------------------------------*/

#if	HOSTWDSZ > 32
#define	Mask32(x)	((x) & 0xffffffff)
#else
#define	Mask32(x)	(x)
#endif


/*---------------------------------------------------------------------------
 * SignExtend$32bits - Force the high-order bits in machines with wordsize
 *  longer than 32 to match bit 0.
 *---------------------------------------------------------------------------*/

#if	HOSTWDSZ > 32
#define	SignEx32(x)	(((x) & 0x80000000) ? ((x) | ((unsigned)-1 << 32)) \
					    : Mask32(x))
#else
#define	SignEx32(x)	(x)
#endif

/**************************/
/* bit field declarations */
/**************************/

/* since the compiler complains if a structure name is declared twice, even
 *  if the declarations are identical, all DeclBitfR invocations are
 *  given here in one file. */

DeclBitfR(0,1,_b01)
DeclBitfR(0,15,_b015)
DeclBitfR(0,16,_b016)
DeclBitfR(0,4,_b04)
DeclBitfR(0,6,_b06)
DeclBitfR(0,8,_b08)
DeclBitfR(4,1,_b41)
DeclBitfR(4,4,_b44)
DeclBitfR(6,1,_b61)
DeclBitfR(6,13,_b613)
DeclBitfR(6,15,_b615)
DeclBitfR(6,17,_b617)
DeclBitfR(6,26,_b626)
DeclBitfR(6,5,_b65)
DeclBitfR(7,1,_b71)
DeclBitfR(8,1,_b81)
DeclBitfR(8,4,_b84)
DeclBitfR(8,8,_b88)
DeclBitfR(9,1,_b91)
DeclBitfR(10,1,_b101)
DeclBitfR(11,1,_b111)
DeclBitfR(11,10,_b1110)
DeclBitfR(11,4,_b114)
DeclBitfR(11,5,_b115)
DeclBitfR(12,1,_b121)
DeclBitfR(12,4,_b124)
DeclBitfR(13,1,_b131)
DeclBitfR(14,1,_b141)
DeclBitfR(15,1,_b151)
DeclBitfR(16,1,_b161)
DeclBitfR(16,15,_b1615)
DeclBitfR(16,16,_b1616)
DeclBitfR(16,2,_b162)
DeclBitfR(16,3,_b163)
DeclBitfR(16,4,_b164)
DeclBitfR(16,5,_b165)
DeclBitfR(16,8,_b168)
DeclBitfR(17,1,_b171)
DeclBitfR(18,1,_b181)
DeclBitfR(18,13,_b1813)
DeclBitfR(18,2,_b182)
DeclBitfR(18,7,_b187)
DeclBitfR(19,1,_b191)
DeclBitfR(19,8,_b198)
DeclBitfR(19,10,_b1910)
DeclBitfR(20,11,_b2011)
DeclBitfR(20,2,_b202)
DeclBitfR(20,4,_b204)
DeclBitfR(21,10,_b2110)
DeclBitfR(21,2,_b212)
DeclBitfR(21,5,_b215)
DeclBitfR(22,5,_b225)
DeclBitfR(23,3,_b233)
DeclBitfR(24,1,_b241)
DeclBitfR(24,4,_b244)
DeclBitfR(24,8,_b248)
DeclBitfR(25,1,_b251)
DeclBitfR(26,1,_b261)
DeclBitfR(27,1,_b271)
DeclBitfR(27,4,_b274)
DeclBitfR(27,5,_b275)
DeclBitfR(28,1,_b281)
DeclBitfR(28,4,_b284)
DeclBitfR(29,1,_b291)
DeclBitfR(30,1,_b301)
DeclBitfR(30,2,_b302)
DeclBitfR(31,1,_b311)

/******************/
/* Word subfields */
/******************/

#define	Sign(i)		BitfR(i,0,1,_b01)
/* halfwords */
#define	Hwd0(i)		BitfR(i,0,16,_b016)
#define	Hwd1sign(i)	BitfR(i,16,1,_b161)
#define	Hwd1(i)		BitfR(i,16,16,_b1616)
/* bytes */
#define	Byte0(i)	BitfR(i,0,8,_b08)
#define	Byte1sign(i)	BitfR(i,8,1,_b81)
#define	Byte1(i)	BitfR(i,8,8,_b88)
#define	Byte2(i)	BitfR(i,16,8,_b168)
#define	Byte3sign(i)	BitfR(i,24,1,_b241)
#define	Byte3(i)	BitfR(i,24,8,_b248)
/* digits */
#define	Digit0(i)	BitfR(i,0,4,_b04)
#define	Digit1(i)	BitfR(i,4,4,_b44)
#define	Digit2(i)	BitfR(i,8,4,_b84)
#define	Digit3(i)	BitfR(i,12,4,_b124)
#define	Digit4(i)	BitfR(i,16,4,_b164)
#define	Digit5(i)	BitfR(i,20,4,_b204)
#define	Digit6(i)	BitfR(i,24,4,_b244)
#define	Digit7(i)	BitfR(i,28,4,_b284)

/* Wordsize definitions */

#define		BIT_P_DW	64	/* bits/doubleword */
#define		BIT_P_WD	32	/* bits/word */
#define		BIT_P_HW	16	/* bits/halfword */
#define		BIT_P_BYT	8	/* bits/byte */
#define		BYT_P_DW	8	/* bytes/doubleword */
#define		BYT_P_WD	4	/* bytes/word */
#define		BYT_P_HW	2	/* bytes/halfword */

/* Masks */

#define		WDMASK		0xffffffff	/* 32-bit mask */
#define		OFSMASK		0xffffffff	/* 32-bit mask */
#define		SIDMASK		0xffffffff	/* 32-bit mask */
#define		SIGNMASK	0x80000000	/* 32 bit word sign bit */

/* Alignments */

#define		wdalign(ofs)	(ofs &= ~3)
/*
 *  $Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel_unused/parisc/kdb/unasm.c,v 1.5 1994/07/21 22:32:05 mike Exp $
 *
 *  Spectrum Simulator Instruction Opcode Definitions
 *  Dan Magenheimer
 *  Computer Research Center, Hewlett-Packard Labs
 *
 *  (c) copyright 1982
 *  (p) protected 1982
 *  The Hewlett-Packard Company
 *  Hewlett-Packard Laboratories
 *  Computer Research Center
 *  Palo Alto, California
 *
 *  *** HP Company Confidential ***
 *
 * $Log: unasm.c,v $
 * Revision 1.5  1994/07/21  22:32:05  mike
 * official HP copyright notice
 *
 * Revision 1.4  1992/07/08  12:19:52  dalton
 * Checkin before split to 1.0.4 release (by LBS).
 *
 * Revision 1.3  92/06/06  16:16:45  dalton
 * *** empty log message ***
 * 
 * Revision 1.2  92/06/06  15:42:28  dalton
 * Changed include to be a path relative to hp800.
 * 
 * Revision 1.1  92/06/06  14:05:33  dalton
 * Initial revision
 * 
 * Revision 1.2  91/04/14  20:29:49  osfrcs
 * 	Initial version.
 * 	[91/03/30  09:20:34  brezak]
 * 
 * Revision 1.1.2.2  91/04/02  10:42:50  brezak
 * 	Initial version.
 * 	[91/03/30  09:20:34  brezak]
 * 
 * Revision 1.1.1.2  91/03/30  09:20:34  brezak
 * 	Initial version.
 * 
 * Revision 6.1  89/09/06  10:39:58  burroughs
 * Added shadow registers for gr0-gr7.
 *     gr0-7 are copied into sh0-7 whenever a trap occurs
 *     the instruction RFIR restores gr0-7 from sh0-7 and returns from
 *     interrupt.
 *     the "sh" command displays the shadow registers
 *     = sh7 0x789 works, too.
 * 
 * Revision 6.0  89/09/01  15:46:37  15:46:37  burroughs (Greg Burroughs)
 * baseline for pcx simple offsite
 * 
 * Revision 5.2  87/09/02  14:30:23  14:30:23  kent
 * separated stat gathering for indexed vs short.
 * this will NOT work if cache hints ever get used
 * since this field was assumed always zero
 * 
 * Revision 5.1  87/02/27  11:12:16  11:12:16  kent (Kent McMullen)
 * update all src to 5.1
 * 
 * Revision 5.0  87/02/18  16:31:35  16:31:35  kent (Kent McMullen)
 * update all revision numbers to 5.0 for release
 * 
 * Revision 1.1  86/07/15  08:34:57  08:34:57  kent (Kent McMullen)
 * Initial revision
 * 
 * Revision 4.1  83/10/25  17:02:34  17:02:34  djm (Daniel J Magenheimer)
 * First release for ACD v4
 * 
 * Revision 3.0  83/06/13  10:24:45  djm (Daniel Magenheimer)
 * First release for distribution
 * 
 *
 */

/*
 * Changes:
 *   01/30/90 ejf Simplify SPOPn support, now only gives assist emulation trap.
 *   01/19/90 ejf Replace linpak instructions with just FSTQ[SX].
 *   12/19/89 ejf Add PA89 new floating point opcode 0E.
 *   12/18/89 ejf Change 5 ops to PA89 format.
 *   12/01/89 ejf Move additional instructions fmas, fmaa, fld2, fst2 to ssILst
 *   09/22/89 ejf Fix unbalanced comments.
 */


/* ..and modified by hand to remove the load/store short references */
/* ..and modified by hand to make memory management ops conform to the
 *   requirement that all subops of a major opcode begin in the same
 *   place and have the same length */

#define	LDW	0x12, 0x00, 0, 0	/* LOAD WORD */
#define	LDWM	0x13, 0x00, 0, 0	/* LOAD WORD and MODIFY */
#define	LDH	0x11, 0x00, 0, 0	/* LOAD HALFWORD */
#define	LDB	0x10, 0x00, 0, 0	/* LOAD BYTE */
#define	LDO	0x0d, 0x00, 0, 0	/* LOAD OFFSET */
#define	STW	0x1a, 0x00, 0, 0	/* STORE WORD */
#define	STWM	0x1b, 0x00, 0, 0	/* STORE WORD and MODIFY */
#define	STH	0x19, 0x00, 0, 0	/* STORE HALFWORD */
#define	STB	0x18, 0x00, 0, 0	/* STORE BYTE */
#define	LDWX	0x03, 0x02, 19, 7	/* LOAD WORD INDEXED */
#define	LDHX	0x03, 0x01, 19, 7	/* LOAD HALFWORD INDEXED */
#define	LDBX	0x03, 0x00, 19, 7	/* LOAD BYTE INDEXED */
#define	LDWAX	0x03, 0x06, 19, 7	/* LOAD WORD ABSOLUTE INDEXED */
#define	LDCWX	0x03, 0x07, 19, 7	/* LOAD and CLEAR WORD INDEXED */
#define LDWS	0x03, 0x42, 19, 7	/* LOAD WORD SHORT DISP */
#define LDHS	0x03, 0x41, 19, 7	/* LOAD HALFWORD SHORT DISP */
#define LDBS	0x03, 0x40, 19, 7	/* LOAD BYTE SHORT DISP */
#define LDWAS	0x03, 0x46, 19, 7	/* LOAD WORD ABSOLUTE SHORT DISP */
#define LDCWS	0x03, 0x47, 19, 7	/* LOAD and CLEAR WORD SHORT DISP */
#define	STWS	0x03, 0x4a, 19, 7	/* STORE WORD SHORT DISP */
#define	STHS	0x03, 0x49, 19, 7	/* STORE HALFWORD SHORT DISP */
#define	STBS	0x03, 0x48, 19, 7	/* STORE BYTE SHORT DISP */
#define	STWAS	0x03, 0x4e, 19, 7	/* STORE WORD ABSOLUTE SHORT DISP */
#define	STBYS	0x03, 0x4c, 19, 7	/* STORE BYTES SHORT DISP */
#define	LDIL	0x08, 0x00, 0, 0	/* LOAD IMMED LEFT */
#define	ADDIL	0x0a, 0x00, 0, 0	/* ADD IMMED LEFT */
#define	BL	0x3a, 0x00, 16, 3	/* BRANCH [and LINK] */
#define	GATE	0x3a, 0x01, 16, 3	/* GATEWAY */
#define	BLR	0x3a, 0x02, 16, 3	/* BRANCH and LINK REGISTER */
#define	BV	0x3a, 0x06, 16, 3	/* BRANCH VECTORED */
#define	BE	0x38, 0x00, 0, 0	/* BRANCH EXTERNAL */
#define	BLE	0x39, 0x00, 0, 0	/* BRANCH and LINK EXTERNAL */
#define	MOVB	0x32, 0x00, 0, 0	/* MOVE and BRANCH */
#define	MOVIB	0x33, 0x00, 0, 0	/* MOVE IMMED and BRANCH */
#define	COMBT	0x20, 0x00, 0, 0	/* COMPARE and BRANCH if TRUE */
#define	COMBF	0x22, 0x00, 0, 0	/* COMPARE and BRANCH if FALSE */
#define	COMIBT	0x21, 0x00, 0, 0	/* COMPARE IMMED and BRANCH if TRUE */
#define	COMIBF	0x23, 0x00, 0, 0	/* COMPARE IMMED and BRANCH if FALSE */
#define	ADDBT	0x28, 0x00, 0, 0	/* ADD and BRANCH if TRUE */
#define	ADDBF	0x2a, 0x00, 0, 0	/* ADD and BRANCH if FALSE */
#define	ADDIBT	0x29, 0x00, 0, 0	/* ADD IMMED and BRANCH if TRUE */
#define	ADDIBF	0x2b, 0x00, 0, 0	/* ADD IMMED and BRANCH if FALSE */
#define	BVB	0x30, 0x00, 0, 0	/* BRANCH on VARIABLE BIT */
#define	BB	0x31, 0x00, 0, 0	/* BRANCH on BIT */
#define	ADD	0x02, 0x30, 20, 7	/* ADD  */
#define	ADDL	0x02, 0x50, 20, 7	/* ADD LOGICAL */
#define	ADDO	0x02, 0x70, 20, 7	/* ADD and TRAP on OVFLO */
#define	SH1ADD	0x02, 0x32, 20, 7	/* SHIFT 1, ADD  */
#define	SH1ADDL	0x02, 0x52, 20, 7	/* SHIFT 1, ADD LOGICAL */
#define	SH1ADDO	0x02, 0x72, 20, 7	/* SHIFT 1, ADD and TRAP on OVFLO */
#define	SH2ADD	0x02, 0x34, 20, 7	/* SHIFT 2, ADD  */
#define	SH2ADDL	0x02, 0x54, 20, 7	/* SHIFT 2, ADD LOGICAL */
#define	SH2ADDO	0x02, 0x74, 20, 7	/* SHIFT 2, ADD and TRAP on OVFLO */
#define	SH3ADD	0x02, 0x36, 20, 7	/* SHIFT 3, ADD  */
#define	SH3ADDL	0x02, 0x56, 20, 7	/* SHIFT 3, ADD LOGICAL */
#define	SH3ADDO	0x02, 0x76, 20, 7	/* SHIFT 3, ADD and TRAP on OVFLO */
#define	ADDC	0x02, 0x38, 20, 7	/* ADD with CARRY  */
#define	ADDCO	0x02, 0x78, 20, 7	/* ADD with CARRY and TRAP on OVFLO */
#define	SUB	0x02, 0x20, 20, 7	/* SUBTRACT  */
#define	SUBO	0x02, 0x60, 20, 7	/* SUBTRACT and TRAP on OVFLO */
#define	SUBB	0x02, 0x28, 20, 7	/* SUBTRACT with BORROW  */
#define	SUBBO	0x02, 0x68, 20, 7	/* SUBTRACT with BORROW and TRAP on OVFLO */
#define	SUBT	0x02, 0x26, 20, 7	/* SUBTRACT and TRAP on COND */
#define	SUBTO	0x02, 0x66, 20, 7	/* SUBTRACT and TRAP on COND or OVFLO */
#define	DS	0x02, 0x22, 20, 7	/* DIVIDE STEP */
#define	COMCLR	0x02, 0x44, 20, 7	/* COMPARE and CLEAR */
#define	OR	0x02, 0x12, 20, 7	/* INCLUSIVE OR */
#define	XOR	0x02, 0x14, 20, 7	/* EXCLUSIVE OR */
#define	AND	0x02, 0x10, 20, 7	/* AND */
#define	ANDCM	0x02, 0x00, 20, 7	/* AND COMPLEMENT */
#define	UXOR	0x02, 0x1c, 20, 7	/* UNIT XOR */
#define	UADDCM	0x02, 0x4c, 20, 7	/* UNIT ADD COMPLEMENT */
#define	UADDCMT	0x02, 0x4e, 20, 7	/* UNIT ADD COMPLEMENT and TRAP on COND */
#define	DCOR	0x02, 0x5c, 20, 7	/* DECIMAL CORRECT */
#define	IDCOR	0x02, 0x5e, 20, 7	/* INTERMEDIATE DECIMAL CORRECT */
#define	ADDI	0x2d, 0x00, 20, 1	/* ADD to IMMED  */
#define	ADDIO	0x2d, 0x01, 20, 1	/* ADD to IMMED and TRAP on OVFLO */
#define	ADDIT	0x2c, 0x00, 20, 1	/* ADD to IMMED and TRAP on COND */
#define	ADDITO	0x2c, 0x01, 20, 1	/* ADD to IMMED and TRAP on COND or OVFLO */
#define	SUBI	0x25, 0x00, 20, 1	/* SUBTRACT from IMMED  */
#define	SUBIO	0x25, 0x01, 20, 1	/* SUBTRACT from IMMED and TRAP on OVFLO */
#define	COMICLR	0x24, 0x00, 0, 0	/* COMPARE IMMED and CLEAR */
#define	VSHD	0x34, 0x00, 19, 3	/* VARIABLE SHIFT DOUBLE */
#define	SHD	0x34, 0x02, 19, 3	/* SHIFT DOUBLE */
#define	VEXTRU	0x34, 0x04, 19, 3	/* VARIABLE EXTRACT RIGHT UNSIGNED */
#define	VEXTRS	0x34, 0x05, 19, 3	/* VARIABLE EXTRACT RIGHT SIGNED */
#define	EXTRU	0x34, 0x06, 19, 3	/* EXTRACT RIGHT UNSIGNED  */
#define	EXTRS	0x34, 0x07, 19, 3	/* EXTRACT RIGHT SIGNED */
#define	VDEP	0x35, 0x01, 19, 3	/* VARIABLE DEPOSIT */
#define	DEP	0x35, 0x03, 19, 3	/* DEPOSIT */
#define	VDEPI	0x35, 0x05, 19, 3	/* VARIABLE DEPOSIT IMMED */
#define	DEPI	0x35, 0x07, 19, 3	/* DEPOSIT IMMED */
#define	ZVDEP	0x35, 0x00, 19, 3	/* ZERO and VARIABLE DEPOSIT */
#define	ZDEP	0x35, 0x02, 19, 3	/* ZERO and DEPOSIT */
#define	ZVDEPI	0x35, 0x04, 19, 3	/* ZERO and VARIABLE DEPOSIT IMMED */
#define	ZDEPI	0x35, 0x06, 19, 3	/* ZERO and DEPOSIT IMMED */
#define	BREAK	0x00, 0x00, 19, 8	/* BREAK */
#define	RFI	0x00, 0x60, 19, 8	/* RETURN FROM INTERRUPTION */
#define	RFIR	0x00, 0x65, 19, 8	/* RFI & RESTORE SHADOW REGISTERS */
#define	SSM	0x00, 0x6b, 19, 8	/* SET SYSTEM MASK */
#define	RSM	0x00, 0x73, 19, 8	/* RESET SYSTEM MASK */
#define	MTSM	0x00, 0xc3, 19, 8	/* MOVE TO SYSTEM MASK */
#define	LDSID	0x00, 0x85, 19, 8	/* LOAD SPACE IDENTIFIER */
#define	MTSP	0x00, 0xc1, 19, 8	/* MOVE TO SPACE REGISTER */
#define	MTCTL	0x00, 0xc2, 19, 8	/* MOVE TO SYSTEM CONTROL REGISTER */
#define	MFSP	0x00, 0x25, 19, 8	/* MOVE FROM SPACE REGISTER */
#define	MFCTL	0x00, 0x45, 19, 8	/* MOVE FROM SYSTEM CONTROL REGISTER */
#define	SYNC	0x00, 0x20, 19, 8	/* SYNCHRONIZE DATA CACHE */
#define	DIAG	0x05, 0x00, 0, 0	/* DIAGNOSE */
#define	SPOP	0x04, 0x00, 0, 0	/* SPECIAL FUNCTION UNIT */
#define	COPR	0x0c, 0x00, 0, 0	/* COPROCESSOR */
#define	CLDWX	0x09, 0x00, 19, 4	/* COPROCESSOR LOAD WORD INDEXED */
#define	CLDDX	0x0b, 0x00, 19, 4	/* COPROCESSOR LOAD WORD INDEXED */
#define	CSTWX	0x09, 0x01, 19, 4	/* COPROCESSOR STORE WORD INDEXED */
#define	CSTDX	0x0b, 0x01, 19, 4	/* COPROCESSOR STORE WORD INDEXED */
#define CLDWS	0x09, 0x08, 19, 4	/* COPROCESSOR LOAD WORD SHORT */
#define CLDDS	0x0b, 0x08, 19, 4	/* COPROCESSOR LOAD WORD SHORT */
#define CSTWS	0x09, 0x09, 19, 4	/* COPROCESSOR STORE WORD SHORT */
#define CSTDS	0x0b, 0x09, 19, 4	/* COPROCESSOR STORE WORD SHORT */
#define	FLOAT0	0x0e, 0x00, 21, 2	/* FLOATING POINT CLASS 0 */
#define	FLOAT1	0x0e, 0x01, 21, 2	/* FLOATING POINT CLASS 1 */
#define	FLOAT2	0x0e, 0x02, 21, 2	/* FLOATING POINT CLASS 2 */
#define	FLOAT3	0x0e, 0x03, 21, 2	/* FLOATING POINT CLASS 3 */
#define	FMPYSUB	0x26, 0x00, 0, 0	/* FP MULTIPLY AND SUBTRACT */
#define	FMPYADD	0x06, 0x00, 0, 0	/* FP MULTIPLY AND ADD/TRUNCATE */
#define	FSTQX	0x0f, 0x01, 19, 4	/* FLOATING POINT STORE QUAD INDEXED */
#define FSTQS	0x0f, 0x09, 19, 4	/* FLOATING POINT STORE QUAD SHORT */
/* all of the following have been pushed around to conform */
#define	PROBER	0x01, 0x46, 19, 7	/* PROBE READ ACCESS */
#ifdef notdef
#define PROBERI 0x01, 0xc6, 19, 7	/* PROBE READ ACCESS IMMEDIATE */
#endif
#define	PROBEW	0x01, 0x47, 19, 7	/* PROBE WRITE ACCESS */
#ifdef notdef
#define PROBEWI 0x01, 0xc7, 19, 7	/* PROBE WRITE ACCESS IMMEDIATE */
#endif
#define	LPA	0x01, 0x4d, 19, 7	/* LOAD PHYSICAL ADDRESS */
#define	LHA	0x01, 0x4c, 19, 7	/* LOAD HASH ADDRESS */
#define	PDTLB	0x01, 0x48, 19, 7	/* PURGE DATA TRANS LOOKASIDE BUFFER */
#define	PITLB	0x01, 0x08, 19, 7	/* PURGE INST TRANS LOOKASIDE BUFFER */
#define	PDTLBE	0x01, 0x49, 19, 7	/* PURGE DATA TLB ENTRY */
#define	PITLBE	0x01, 0x09, 19, 7	/* PURGE INST TLB ENTRY */
#define	IDTLBA	0x01, 0x41, 19, 7	/* INSERT DATA TLB ADDRESS */
#define	IITLBA	0x01, 0x01, 19, 7	/* INSERT INSTRUCTION TLB ADDRESS */
#define	IDTLBP	0x01, 0x40, 19, 7	/* INSERT DATA TLB PROTECTION */
#define	IITLBP	0x01, 0x00, 19, 7	/* INSERT INSTRUCTION TLB PROTECTION */
#define	PDC	0x01, 0x4e, 19, 7	/* PURGE DATA CACHE */
#define	FDC	0x01, 0x4a, 19, 7	/* FLUSH DATA CACHE */
#define	FIC	0x01, 0x0a, 19, 7	/* FLUSH INSTRUCTION CACHE */
#define	FDCE	0x01, 0x4b, 19, 7	/* FLUSH DATA CACHE ENTRY */
#define	FICE	0x01, 0x0b, 19, 7	/* FLUSH DATA CACHE ENTRY */

/*
 *  $Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel_unused/parisc/kdb/unasm.c,v 1.5 1994/07/21 22:32:05 mike Exp $
 *
 *  Spectrum Simulator Instruction Set Constants and Datatypes
 *  Dan Magenheimer - 4/28/82
 *  Computer Research Center, Hewlett-Packard Labs
 *
 *  (c) copyright 1982
 *  (p) protected 1982
 *  The Hewlett-Packard Company
 *  Hewlett-Packard Laboratories
 *  Computer Research Center
 *  Palo Alto, California
 *
 *  *** HP Company Confidential ***
 *
 * $Log: unasm.c,v $
 * Revision 1.5  1994/07/21  22:32:05  mike
 * official HP copyright notice
 *
 * Revision 1.4  1992/07/08  12:19:52  dalton
 * Checkin before split to 1.0.4 release (by LBS).
 *
 * Revision 1.3  92/06/06  16:16:45  dalton
 * *** empty log message ***
 * 
 * Revision 1.2  92/06/06  15:42:28  dalton
 * Changed include to be a path relative to hp800.
 * 
 * Revision 1.1  92/06/06  14:05:33  dalton
 * Initial revision
 * 
 * Revision 1.2  91/04/14  20:29:49  osfrcs
 * 	Initial version.
 * 	[91/03/30  09:20:34  brezak]
 * 
 * Revision 1.1.2.2  91/04/02  10:42:50  brezak
 * 	Initial version.
 * 	[91/03/30  09:20:34  brezak]
 * 
 * Revision 1.1.1.2  91/03/30  09:20:34  brezak
 * 	Initial version.
 * 
;Revision 1.1  88/07/11  14:05:21  14:05:21  ren (Bob Naas)
;Initial revision
;
 * Revision 5.1  87/02/27  11:12:23  11:12:23  kent (Kent McMullen)
 * update all src to 5.1
 * 
 * Revision 5.0  87/02/18  16:31:52  16:31:52  kent (Kent McMullen)
 * update all revision numbers to 5.0 for release
 * 
 * Revision 1.1  86/07/15  08:35:00  08:35:00  kent (Kent McMullen)
 * Initial revision
 * 
 * Revision 4.3  85/11/12  09:28:44  09:28:44  viggy (Viggy Mokkarala)
 * first mpsim version, partially stable
 * 
 * Revision 4.2  84/07/16  17:20:57  17:20:57  djm ()
 * Define field macros for COPR and SFU insts
 * 
 * Revision 4.1  83/10/25  17:10:14  djm (Daniel Magenheimer)
 * First release for ACD v4
 * 
 * Revision 3.1  83/08/03  14:09:59  djm (Daniel Magenheimer)
 * Sys calls, args, -S, bug fixes, etc.
 * 
 * Revision 3.0  83/06/13  10:25:13  djm (Daniel Magenheimer)
 * First release for distribution
 * 
 *
 */
/*
 * Changes:
 *   12/01/89 ejf Add Rsd(), Rse(), Rtd(), Rte() for 5 ops.
 *   11/30/89 ejf Make instruction use counters shared, not per cpu.
 *   11/28/89 ejf Change majoropcode for quicker extension extract.
 */



/*
 *  Dependencies: std.h, ssDefs.h, bits.h
 */


/* Lookup/Execute structure for instructions */
struct inst {
	uint	majopc;		/* major opcode of instruction, 0..MAXOPC */
	uint	opcext;		/* opcode extension, 0 if not applic. */
	uint	extbs;		/* starting bit pos of extension field */
	uint	extbl;		/* bit length of extension field */
	uint	count;		/* frequency counter for analysis */
	char	mnem[8];	/* ascii mnemonic */
	void	*dasmfcn;	/* disassembly function */
};


#define	NMAJOPCS	64



struct majoropcode {
	struct inst **subops; /* pointer to table of subops indexed by
				   *  opcode extension */
	uint	maxsubop;	/* largest opcode extension value or 0 */
	uint	extshft;	/* right shift amount for extension field */
	uint	extmask;	/* post shift mask for extension field */
};

#define	OpExt(i,m)	((i >> m->extshft) & m->extmask)	/* extract opcode extension */


/*****************************/
/* Miscellaneous definitions */
/*****************************/

/* Load/Store Indexed Opcode Extension Cache Control */
#define	NOACTION	0
#define	STACKREF	1
#define	SEQPASS		2
#define	PREFETCH	3

/******************************/
/* Fields within instructions */
/******************************/

/* opcode */
#define	Opcode(i)	BitfR(i,0,6,_b06)
/* opcode true/false bit */
#define	OpcTF(i)	BitfR(i,4,1,_b41)
/* register sources */
#define	Rsa(i)		BitfR(i,11,5,_b115)
#define	Rsb(i)		BitfR(i,6,5,_b65)
#define	Rsc(i)		BitfR(i,27,5,_b275)
#define	Rsd(i)		BitfR(i,21,5,_b215)
#define	Rse(i)		BitfR(i,16,5,_b165)
/* register targets */
#define	Rta(i)		BitfR(i,11,5,_b115)
#define	Rtb(i)		BitfR(i,6,5,_b65)
#define	Rtc(i)		BitfR(i,27,5,_b275)
#define	Rtd(i)		BitfR(i,21,5,_b215)
#define	Rte(i)		BitfR(i,16,5,_b165)
/* 5-bit immediates (Magnitude, Sign) */
#define	Imb5(i)		BitfR(i,6,5,_b65)
#define	Ima5M(i)	BitfR(i,11,4,_b114)
#define	Ima5S(i)	BitfR(i,15,1,_b151)
#define	Ima5A(i)	BitfR(i,11,5,_b115)
#define	Imd5(i)		BitfR(i,22,5,_b225)
#define	Imc5M(i)	BitfR(i,27,4,_b274)
#define	Imc5S(i)	BitfR(i,31,1,_b311)
#define	Imc5A(i)	BitfR(i,27,5,_b275)
/* Other immediates */
#define	Im21L(i)	BitfR(i,18,2,_b182)
#define	Im21H(i)	BitfR(i,20,11,_b2011)
#define	Im21M1(i)	BitfR(i,16,2,_b162)
#define	Im21M2(i)	BitfR(i,11,5,_b115)
#define	Im21S(i)	BitfR(i,31,1,_b311)
#define	Im11M(i)	BitfR(i,21,10,_b2110)
#define	Im11S(i)	BitfR(i,31,1,_b311)
/* displacements/offsets */
#define	DispM(i)	BitfR(i,18,13,_b1813)
#define	DispS(i)	BitfR(i,31,1,_b311)
#define	Off5(i)		BitfR(i,11,5,_b115)
#define	Off11H(i)	BitfR(i,19,10,_b1910)
#define	Off11L(i)	BitfR(i,29,1,_b291)
#define	OffS(i)		BitfR(i,31,1,_b311)
/* miscellaneous */
#define	Dss(i)		BitfR(i,16,2,_b162)
#define	Cond(i)		BitfR(i,16,3,_b163)
#define	Cneg(i)		BitfR(i,19,1,_b191)
#define	Cond4(i)	BitfR(i,16,4,_b164)	/* Cond AND Cneg */
#define	Nu(i)		BitfR(i,30,1,_b301)
#define	SrL(i)		BitfR(i,16,2,_b162)
#define	SrH(i)		BitfR(i,18,1,_b181)
#define	ShortDisp(i)	BitfR(i,19,1,_b191)
#define	IndxShft(i)	BitfR(i,18,1,_b181)
#define	ModBefore(i)	BitfR(i,18,1,_b181)
#define	CacheCtrl(i)	BitfR(i,20,2,_b202)
#define	Modify(i)	BitfR(i,26,1,_b261)
#define	ProbeI(i)	BitfR(i,18,1,_b181)
#define	Uid(i)		BitfR(i,23,3,_b233)
#define	Sfu(i)		BitfR(i,23,3,_b233)
#define	CopExt17(i)	BitfR(i,6,17,_b617)
#define	CopExt5(i)	BitfR(i,27,5,_b275)
#define	SpopType(i)	BitfR(i,21,2,_b212)
#define	SpopExt15(i)	BitfR(i,6,15,_b615)
#define	SpopExt10(i)	BitfR(i,11,10,_b1110)
#define	SpopExt5L(i)	BitfR(i,16,5,_b165)
#define	SpopExt5(i)	BitfR(i,27,5,_b275)
#define	NoMajOpc(i)	BitfR(i,6,26,_b626)
#define	Bi1(i)		BitfR(i,27,5,_b275)	/* fields in BREAK */
#define	Bi2(i)		BitfR(i,6,13,_b613)

/* fragmented field collating macros */
#define	Ima5(i)		(Ima5S(i) ? Ima5M(i) | (-1<<4) : Ima5M(i))

#define	Imc5(i)		(Imc5S(i) ? Imc5M(i) | (-1<<4) : Imc5M(i))

#define	Disp(i)		(DispS(i) ?   DispM(i) | (-1<<13) : DispM(i))

#define	Im21(i)		(Im21S(i) << 31 | Im21H(i) << 20 | Im21M1(i) << 18 | \
				Im21M2(i) << 13 | Im21L(i) << 11)

#define	Im11(i)		(Im11S(i) ?   Im11M(i) | (-1<<10) : Im11M(i))

#define	Bdisp(i)	((OffS(i) ? (Off5(i)<<11 | Off11L(i)<<10|Off11H(i)) \
/* branch displacement (bytes) */	| (-1 << 16)			\
				  : (Off5(i)<<11|Off11L(i)<<10|Off11H(i))) << 2)

#define	Cbdisp(i)	((OffS(i) ?   (Off11L(i) << 10 | Off11H(i)) \
 /* compare/branch disp (bytes) */ | (-1 << 11)			\
				  :    Off11L(i) << 10 | Off11H(i)) << 2)

#define	Sr(i)		(SrH(i)<<2 | SrL(i))

/* sfu/copr */
#define	CoprExt1(i)	(CopExt17(i))
#define	CoprExt2(i)	(CopExt5(i))
#define	CoprExt(i)	((CopExt17(i)<<5) | CopExt5(i))
#define	Spop0Ext(i)	((SpopExt15(i)<<5) | SpopExt5(i))
#define	Spop1Ext(i)	(SpopExt15(i))
#define	Spop2Ext(i)	((SpopExt10(i)<<5) | SpopExt5(i))
#define	Spop3Ext(i)	((SpopExt5L(i)<<5) | SpopExt5(i))


static int
strncmp(s1, s2, n)
register unsigned char		*s1;
register unsigned char		*s2;
register int			n;
{
	do {
		if (*s1 != *s2++)
			return(*(char *)s1 - *(char *)--s2);
	} while (--n != 0 && *s1++ != '\0');

	return(0);
}

/*##################### Globals - Imports ##################################*/

/* Disassembly functions */
/* Any instr disassembly replacement routines must be IMPORTed here */

IMPORT BOOL	ldDasm(), stDasm(), ldxDasm(), stsDasm(), stbysDasm();
IMPORT BOOL	ariDasm(), addDasm(), unitDasm();
IMPORT BOOL	limmDasm();
IMPORT BOOL	brDasm(), bvDasm(), beDasm();
IMPORT BOOL	blDasm(), cbDasm(), cbiDasm(), bbDasm();
IMPORT BOOL	iaDasm(), extrDasm(), depDasm(), depiDasm();
IMPORT BOOL	shdDasm(), vextrDasm(), vdepDasm(), vdepiDasm();
IMPORT BOOL	mmgtDasm(), scDasm();
IMPORT BOOL	coprDasm(), cldDasm(), cstDasm();
IMPORT BOOL	brkDasm();
IMPORT BOOL	diagDasm();
IMPORT BOOL	floatDasm(), fmpysubDasm(), fmpyaddDasm(), lpkDasm();

/*##################### Globals - Exports ##################################*/

EXPORT struct majoropcode majopcs[NMAJOPCS] = {0};

EXPORT struct inst illeg = { 0, 0, 0, 0, 0, "???", 0 };

/* To replace instr function, do the following:				*/
/*	a) locate the desired entry in instrs[] below			*/
/*	b) change the 3rd field if an alternate mneumonic is 		*/
/*	   desired for window disassembly				*/
/*	c) change the 4th field to the name of the function being	*/
/* 	   used for replacement (i.e. ldwRepl instead of ldw)		*/
/*	d) change the 5th field if an alternate disassembly routine	*/
/*	   is desired (i.e. ldDasmRepl)					*/

EXPORT struct inst instrs[] = {
	{ LDW,    0, "LDW",	ldDasm },
	{ LDH,    0, "LDH",	ldDasm },
	{ LDB,    0, "LDB",	ldDasm },
	{ LDWM,   0, "LDWM",    ldDasm },
	{ LDO,    0, "LDO",     ldDasm },
	{ STW,    0, "STW",     stDasm },
	{ STH,    0, "STH",     stDasm },
	{ STB,    0, "STB",     stDasm },
	{ STWM,   0, "STWM",    stDasm },
	{ LDWX,   0, "LDW",	ldxDasm },
	{ LDHX,   0, "LDH",	ldxDasm },
	{ LDBX,   0, "LDB",	ldxDasm },
	{ LDCWX,  0, "LDCW",	ldxDasm },
	{ LDWAX,  0, "LDWA",	ldxDasm },
	{ LDWS,   0, "LDW",	ldxDasm },
	{ LDHS,   0, "LDH",	ldxDasm },
	{ LDBS,   0, "LDB",	ldxDasm },
	{ LDCWS,  0, "LDCW",	ldxDasm },
	{ LDWAS,  0, "LDWA",	ldxDasm },
	{ STWS,   0, "STWS",    stsDasm },
	{ STHS,   0, "STHS",    stsDasm },
	{ STBS,   0, "STBS",    stsDasm },
	{ STWAS,  0, "STWAS",   stsDasm },
	{ STBYS,  0, "STBYS",   stbysDasm },
	{ LDIL,   0, "LDIL",    limmDasm },
	{ ADDIL,  0, "ADDIL",   limmDasm },
	{ GATE,   0, "GATE",    blDasm },
	{ BL,     0, "B",	blDasm },
	{ BLR,    0, "BLR",     brDasm },
	{ BV,     0, "BV",      bvDasm },
	{ BE,     0, "BE",      beDasm },
	{ BLE,    0, "BLE",     beDasm },
	{ COMBT,  0, "COMBT",   cbDasm },
	{ COMBF,  0, "COMBF",   cbDasm },
	{ COMIBT, 0, "COMIBT",  cbiDasm },
	{ COMIBF, 0, "COMIBF",  cbiDasm },
	{ ADDBT,  0, "ADDBT",   cbDasm },
	{ ADDBF,  0, "ADDBF",   cbDasm },
	{ ADDIBT, 0, "ADDIBT",  cbiDasm },
	{ ADDIBF, 0, "ADDIBF",  cbiDasm },
	{ MOVB,   0, "MOVB",    cbDasm },
	{ MOVIB,  0, "MOVIB",   cbiDasm },
	{ BB,     0, "BB",      bbDasm },
	{ BVB,    0, "BVB",     bbDasm },
	{ SUBO,   0, "SUBO",    ariDasm },
	{ ADD,    0, "ADD",     addDasm },
	{ ADDL,   0, "ADDL",    addDasm },
	{ ADDO,   0, "ADDO",    ariDasm },
	{ SH1ADD, 0, "SH1ADD",  ariDasm },
	{ SH1ADDL,0, "SH1ADDL", ariDasm },
	{ SH1ADDO,0, "SH1ADDO", ariDasm },
	{ SH2ADD, 0, "SH2ADD",  ariDasm },
	{ SH2ADDL,0, "SH2ADDL", ariDasm },
	{ SH2ADDO,0, "SH2ADDO", ariDasm },
	{ SH3ADD, 0, "SH3ADD",  ariDasm },
	{ SH3ADDL,0, "SH3ADDL", ariDasm },
	{ SH3ADDO,0, "SH3ADDO", ariDasm },
	{ SUB,    0, "SUB",     ariDasm },
	{ ADDCO,  0, "ADDCO",   ariDasm },
	{ SUBBO,  0, "SUBBO",   ariDasm },
	{ ADDC,   0, "ADDC",    ariDasm },
	{ SUBB,   0, "SUBB",    ariDasm },
	{ COMCLR, 0, "COMCLR",  ariDasm },
	{ OR,     0, "OR",      ariDasm },
	{ AND,    0, "AND",     ariDasm },
	{ XOR,    0, "XOR",     ariDasm },
	{ ANDCM,  0, "ANDCM",   ariDasm },
	{ DS,     0, "DS",      ariDasm },
	{ UXOR,   0, "UXOR",    unitDasm },
	{ UADDCM, 0, "UADDCM",  unitDasm },
	{ UADDCMT,0, "UADDCMT", unitDasm },
	{ SUBTO,  0, "SUBTO",   ariDasm },
	{ SUBT,   0, "SUBT",    ariDasm },
	{ DCOR,   0, "DCOR",    unitDasm },
	{ IDCOR,  0, "IDCOR",   unitDasm },
	{ ADDIO,  0, "ADDIO",   iaDasm },
	{ SUBIO,  0, "SUBIO",   iaDasm },
	{ ADDI,   0, "ADDI",    iaDasm },
	{ SUBI,   0, "SUBI",    iaDasm },
	{ COMICLR,0, "COMICLR", iaDasm },
	{ ADDITO, 0, "ADDITO",  iaDasm },
	{ ADDIT,  0, "ADDIT",   iaDasm },
	{ SHD,    0, "SHD",     shdDasm },
	{ VSHD,   0, "VSHD",    shdDasm },
	{ EXTRU,  0, "EXTRU",   extrDasm },
	{ EXTRS,  0, "EXTRS",   extrDasm },
	{ VEXTRU, 0, "VEXTRU",  vextrDasm },
	{ VEXTRS, 0, "VEXTRS",  vextrDasm },
	{ DEP,    0, "DEP",     depDasm },
	{ VDEP,   0, "VDEP",    vdepDasm },
	{ DEPI,   0, "DEPI",    depiDasm },
	{ VDEPI,  0, "VDEPI",   vdepiDasm },
	{ ZDEP,   0, "ZDEP",    depDasm },
	{ ZVDEP,  0, "ZVDEP",   vdepDasm },
	{ ZDEPI,  0, "ZDEPI",   depiDasm },
	{ ZVDEPI, 0, "ZVDEPI",  vdepiDasm },
	{ BREAK,  0, "BREAK",   brkDasm },
	{ RFI,    0, "RFI",     0 },
	{ RFIR,   0, "RFIR",    0 },
	{ SSM,    0, "SSM",     scDasm },
	{ RSM,    0, "RSM",     scDasm },
	{ MTSM,   0, "MTSM",    scDasm },
	{ PROBER, 0, "PROBER",  mmgtDasm },
	{ PROBEW, 0, "PROBEW",  mmgtDasm },
	{ LPA,    0, "LPA",     mmgtDasm },
	{ LHA,    0, "LHA",     mmgtDasm },
	{ LDSID,  0, "LDSID",   scDasm },
	{ PDTLB,  0, "PDTLB",   mmgtDasm },
	{ PDTLBE, 0, "PDTLBE",  mmgtDasm },
	{ PITLB,  0, "PITLB",   mmgtDasm },
	{ PITLBE, 0, "PITLBE",  mmgtDasm },
	{ IDTLBA, 0, "IDTLBA",  mmgtDasm },
	{ IITLBA, 0, "IITLBA",  mmgtDasm },
	{ IDTLBP, 0, "IDTLBP",  mmgtDasm },
	{ IITLBP, 0, "IITLBP",  mmgtDasm },
	{ FIC,    0, "FIC",     mmgtDasm },
	{ FICE,   0, "FICE",    mmgtDasm },
	{ PDC,    0, "PDC",     mmgtDasm },
	{ FDC,    0, "FDC",     mmgtDasm },
	{ FDCE,   0, "FDCE",    mmgtDasm },
	{ SYNC,   0, "SYNC",    0 },
	{ MTSP,   0, "MTSP",    scDasm },
	{ MTCTL,  0, "MTCTL",   scDasm },
	{ MFSP,   0, "MFSP",    scDasm },
	{ MFCTL,  0, "MFCTL",   scDasm },
	{ DIAG,   0, "DIAG",    diagDasm },
	{ SPOP,   0, "???",     0 },
	{ COPR,   0, "COPR",    coprDasm },
	{ CLDWX,  0, "CLDW",    coprDasm },
	{ CLDDX,  0, "CLDD",    coprDasm },
	{ CSTWX,  0, "CSTW",    coprDasm },
	{ CSTDX,  0, "CSTD",    coprDasm },
	{ CLDWS,  0, "CLDW",    coprDasm },
	{ CLDDS,  0, "CLDD",    coprDasm },
	{ CSTWS,  0, "CSTW",    coprDasm },
	{ CSTDS,  0, "CSTD",    coprDasm },
	{ FLOAT0, 0, "F",       floatDasm },
	{ FLOAT1, 0, "FCNV",    floatDasm },
	{ FLOAT2, 0, "F",       floatDasm },
	{ FLOAT3, 0, "F",       floatDasm },
	{ FMPYSUB,0, "FMPY",    fmpysubDasm },
	{ FMPYADD,0, "FMPY",    fmpyaddDasm },
	{ FSTQX,  0, "FSTQX",   lpkDasm  },
	{ FSTQS,  0, "FSTQS",   lpkDasm  },
	{0}
};


int unasm_inited = 0;


/*--------------------------------------------------------------------------
 * instruction$ExecutionInitialize - Initialize the instruction execution
 *  data structures.
 *---------------------------------------------------------------------------*/
iExInit(void)
{
	FAST struct inst *i;
	FAST struct majoropcode *m;
	FAST uint j;
	uint	shft, mask;

	/*
	 * Determine maxsubop for each major opcode.
	 * Also, check all instructions of a given major opcode
	 * for consistent opcode extension field definition, and
	 * save a converted form of this definition in the majopcs
	 * entry for this major opcode.
	 */
	for (i = &instrs[0]; *i->mnem; i++) {
		m = &majopcs[i->majopc];
		if (m->maxsubop < i->opcext) m->maxsubop = i->opcext;
		shft = 32 - i->extbs - i->extbl;
		mask = (1 << i->extbl) - 1;
		if (m->extshft || m->extmask) {
			if (m->extshft != shft || m->extmask != mask) {
				printf("%s - Bad instruction initialization!!\n", i->mnem);
				return(NO);
			}
		}
		else { m->extshft = shft; m->extmask = mask; }
	}
	/*
	 * For each major opcode, allocate a subops table using the length
	 * determined above, and initialize all subops to illegal.
	 */
	for (m = &majopcs[0]; m < &majopcs[NMAJOPCS]; m++) {
		m->subops = (struct inst **)sbrk((m->maxsubop+1)*sizeof(i));
		if (m->subops == NULL) {
			printf("Out of memory in iExInit\n");
			return(NO);
		}
		for (j = 0; j <= m->maxsubop; j++)
			m->subops[j] = &illeg;
	}
	/*
	 * Lastly, fill in all legal subops with the appropriate info.
	 */
	for (i = &instrs[0]; *i->mnem; i++)
		majopcs[i->majopc].subops[i->opcext] = i;
	return(YES);
}



/*##################### Globals - Imports ##################################*/

IMPORT	struct majoropcode majopcs[];
IMPORT	struct inst illeg;

/*##################### Globals - Exports ##################################*/
/*##################### Local Variables ####################################*/

static	WORD	w;	/* temp for saving passed instruction */
static	char	*fcoprUndef = "COPR\t(rsvd or undef.)";
static	char	*fmtStrTbl[] = { "SGL", "DBL", "SGL", "QUAD" };
static	char	*condStrTbl[] = {
	    "false?", "false", "?", "!<=>", "=", "=T", "?=", "!<>",
	    "!?>=", "<", "?<", "!>=", "!?>", "<=", "?<=", "!>",
	    "!?<=", ">", "?>", "!<=", "!?<", ">=", "?>=", "!<",
	    "!?=", "<>", "!=", "!=T", "!?", "<=>", "true?", "true"
};
static	char	*fsreg[] = {
	    "r0L",  "r0R",  "r1L",  "r1R",  "r2L",  "r2R",  "r3L",  "r3R",
	    "r4L",  "r4R",  "r5L",  "r5R",  "r6L",  "r6R",  "r7L",  "r7R",
	    "r8L",  "r8R",  "r9L",  "r9R",  "r10L", "r10R", "r11L", "r11R",
	    "r12L", "r12R", "r13L", "r13R", "r14L", "r14R", "r15L", "r15R",
	    "r16L", "r16R", "r17L", "r17R", "r18L", "r18R", "r19L", "r19R",
	    "r20L", "r20R", "r21L", "r21R", "r22L", "r22R", "r23L", "r23R",
	    "r24L", "r24R", "r25L", "r25R", "r26L", "r26R", "r27L", "r27R",
	    "r28L", "r28R", "r29L", "r29R", "r30L", "r30R", "r31L", "r31R"
};
static	char	*fdreg[] = {
	    "r0",   "r0",   "r1",   "r1",   "r2",   "r2",   "r3",   "r3",
	    "r4",   "r4",   "r5",   "r5",   "r6",   "r6",   "r7",   "r7",
	    "r8",   "r8",   "r9",   "r9",   "r10",  "r10",  "r11",  "r11",
	    "r12",  "r12",  "r13",  "r13",  "r14",  "r14",  "r15",  "r15",
	    "r16",  "r16",  "r17",  "r17",  "r18",  "r18",  "r19",  "r19",
	    "r20",  "r20",  "r21",  "r21",  "r22",  "r22",  "r23",  "r23",
	    "r24",  "r24",  "r25",  "r25",  "r26",  "r26",  "r27",  "r27",
	    "r28",  "r28",  "r29",  "r29",  "r30",  "r30",  "r31",  "r31"
};

/*##################### Macros #############################################*/

#define	Match(s)	(strncmp(s,i->mnem,sizeof(s)-1) == 0)

/* bits for assist ops */
#define	AstNu(w)	Modify(w)
#define	Fpi(w)		(Uid(w)>3)

/* bits for 5 ops */
#define	SinglePrec(i)	Modify(i)
#define	Ms1(i)		((Rsb(i)<<1)+(SinglePrec(i)?((Rsb(i)>15)?1:32):0))
#define	Ms2(i)		((Rsa(i)<<1)+(SinglePrec(i)?((Rsa(i)>15)?1:32):0))
#define	Mt(i)		((Rtc(i)<<1)+(SinglePrec(i)?((Rtc(i)>15)?1:32):0))
#define	As(i)		((Rsd(i)<<1)+(SinglePrec(i)?((Rsd(i)>15)?1:32):0))
#define	Ad(i)		((Rte(i)<<1)+(SinglePrec(i)?((Rte(i)>15)?1:32):0))

/*##################### Functions and Subroutines ##########################*/

/***
BOOL	mmgtDasm(), ariDasm(), ldDasm(), scDasm(), coprDasm(), diagDasm();
****/

/**************************************/
/* Miscellaneous Disassembly Routines */
/**************************************/

/*---------------------------------------------------------------------------
 * conditionType$DisassembleCondition - Return a string which contains the
 *  ascii description of the passed numeric condition.
 *---------------------------------------------------------------------------*/

static TEXT subDCond(FAST uint cond)
{
	switch(cond) {
		case EQZ:	return(",=");
		case LT:	return(",<");
		case LE:	return(",<=");
		case LLT:	return(",<<");
		case LLE:	return(",<<=");
		case SV:	return(",SV");
		case OD:	return(",OD");
		case NEQZ:	return(",<>");
		case GE:	return(",>=");
		case GT:	return(",>");
		case LGE:	return(",>>=");
		case LGT:	return(",>>");
		case NSV:	return(",NSV");
		case EV:	return(",EV");
		case TR:	return(",TR");
		case NEV:	return("");
	}
}


/*---------------------------------------------------------------------------
 * conditionType$DisassembleCondition - Return a string which contains the
 *  ascii description of the passed numeric condition.
 *---------------------------------------------------------------------------*/

static TEXT addDCond(FAST uint cond)
{
	switch(cond) {
		case EQZ:	return(",=");
		case LT:	return(",<");
		case LE:	return(",<=");
		case NUV:	return(",NUV");
		case ZNV:	return(",ZNV");
		case SV:	return(",SV");
		case OD:	return(",OD");
		case NEQZ:	return(",<>");
		case GE:	return(",>=");
		case GT:	return(",>");
		case UV:	return(",UV");
		case VNZ:	return(",VNZ");
		case NSV:	return(",NSV");
		case EV:	return(",EV");
		case TR:	return(",TR");
		case NEV:	return("");
	}
}

static TEXT unitDCond(FAST uint cond)
{
	switch(cond) {
		case SHC:	return(",SHC");
		case SHZ:	return(",SHZ");
		case SBC:	return(",SBC");
		case SBZ:	return(",SBZ");
		case SDC:	return(",SDC");
		case NHC:	return(",NHC");
		case NHZ:	return(",NHZ");
		case NBC:	return(",NBC");
		case NBZ:	return(",NBZ");
		case NDC:	return(",NDC");
		case TR:	return(",TR");
		case NEV:	return("");
	}
}

static TEXT edDCond(FAST uint cond)
{
	switch(cond) {
		case XOD:	return(",OD");
		case XTR:	return(",TR");
		case XNE:	return(",<>");
		case XLT:	return(",<");
		case XEQ:	return(",=");
		case XGE:	return(",>=");
		case XEV:	return(",EV");
		case NEV:	return("");
	}
}



/****************************************/
/* Format Specific Disassembly Routines */
/****************************************/


/* Load [modify] instructions */
BOOL ldDasm(struct inst *i)
{
	FAST uint d = Disp(w);

	if (Rsb(w) == 0 && Match("LDO")) {
		printf("LDI\t%D,%%r%d",d,Rta(w));
		return(YES);
	}
	if (d < 2048)
		printf("%s\tR'%X",i->mnem,d);
	else
		printf("%s\t%D",i->mnem,d);
	if (Dss(w))
		printf("(%%sr%d,%%r%d),%%r%d",Dss(w),Rsb(w),Rta(w));
	else
		printf("(%%r%d),%%r%d",Rsb(w),Rta(w));
	return(YES);
}

/* Store [modify] instructions */
BOOL stDasm(void)
{
	FAST uint d = Disp(w);

	printf("\t%%r%d,",Rta(w));
	if (d < 2048)
		printf("R'%X",d);
	else
		printf("%D",d);
	if (Dss(w))
		printf("(%%sr%d,%%r%d)",Dss(w),Rsb(w));
	else
		printf("(%%r%d)",Rsb(w));
	return(YES);
}

/* Load indexed instructions */
BOOL ldxDasm(void)
{
	if (ShortDisp(w)) {
		printf("S");
		if (Modify(w)) {
			if (ModBefore(w)) printf(",MB");
			else printf(",MA");
		}
	}
	else {
		printf("X");
		if (Modify(w)) {
			if (IndxShft(w)) printf(",SM");
			else printf(",M");
		}
		else if (IndxShft(w)) printf(",S");
	}
	switch (CacheCtrl(w)) {
		case NOACTION:	break;
		case STACKREF:	printf(",C"); break;
		case SEQPASS:	printf(",Q"); break;
		case PREFETCH:	printf(",P"); break;
	}
	if (ShortDisp(w))
		printf("\t%d",Ima5(w));
	else
		printf("\t%%r%d",Rsa(w));
	if (Dss(w))
		printf("(%%sr%d,%%r%d),%%r%d",Dss(w),Rsb(w),Rtc(w));
	else
		printf("(%%r%d),%%r%d",Rsb(w),Rtc(w));
	return(YES);
}

/* Store short displacement instructions */
BOOL stsDasm(void)
{
	if (Modify(w)) {
		if (ModBefore(w)) printf(",MB");
		else printf(",MA");
	}
	switch (CacheCtrl(w)) {
		case NOACTION:	break;
		case STACKREF:	printf(",C"); break;
		case SEQPASS:	printf(",Q"); break;
		case PREFETCH:	printf(",P"); break;
	}
	printf("\t%%r%d,",Rta(w));
	if (Dss(w))
		printf("%d(%%sr%d,%%r%d)",Imc5(w),Dss(w),Rsb(w));
	else
		printf("%d(%%r%d)",Imc5(w),Rsb(w));
	return(YES);
}

/* Store Bytes Instruction */
BOOL stbysDasm(void)
{
	if (ModBefore(w)) printf(",E");
	else printf(",B");
	if (Modify(w)) printf(",M");
	switch (CacheCtrl(w)) {
		case NOACTION:	break;
		case STACKREF:	printf(",F"); break;
		case SEQPASS:	printf(",R"); break;
		case PREFETCH:	printf(",Z"); break;
	}
	printf("\t%%r%d,",Rta(w));
	if (Dss(w))
		printf("%d(%%sr%d,%%r%d)",Imc5(w),Dss(w),Rsb(w));
	else
		printf("%d(%%r%d)",Imc5(w),Rsb(w));
	return(YES);
}

/* Long Immediate instructions */
BOOL limmDasm(void)
{
	FAST uint i = Im21(w);

	printf("\tL'%X,%%r%d", i, Rtb(w));
	return(YES);
}


/* Branch and Link instruction(s) (Branch, too!!) */
BOOL blDasm(struct inst *i,OFS ofs)
{
	FAST OFS tgtofs = ofs + 8 + Bdisp(w);
	FAST uint link = Rtb(w);
	uint rmdr;
	char *buf, *s;

	if (link && !Match("GATE")) printf("L");
	if (Nu(w)) printf(",N");
	printf("\t");
	psymoff(tgtofs, ISYM, "");
	if (link || Match("GATE")) printf(",%%r%d",link);
	return(YES);
}

/* Branch Register instruction */
BOOL brDasm(void)
{
	if (Nu(w)) printf(",N");
	printf("\t%%r%d,%%r%d",Rsa(w),Rtb(w));
	return(YES);
}

/* Dispatch instructions */
BOOL bvDasm(void)
{
	if (Nu(w)) printf(",N");
	printf("\t%%r%d(%%r%d)",Rsa(w),Rsb(w));
	return(YES);
}

/* Branch External instructions */
BOOL beDasm(void)
{
	FAST uint d = Bdisp(w);

	if (Nu(w)) printf(",N");
	if (d < 2048)
		printf("\tR'%X(%%sr%d,%%r%d)",d,Sr(w),Rsb(w));
	else
		printf("\t%D(%%sr%d,%%r%d)",d,Sr(w),Rsb(w));
	return(YES);
}


/* Compare/Add and Branch instructions */
BOOL cbDasm(struct inst *i,OFS ofs)
{
	FAST OFS tgtofs = ofs + 8 + Cbdisp(w);
	uint rmdr;
	char *buf, *s;

	if (Match("MOVB")) printf(edDCond(Cond(w)));
	else if (Match("ADDB")) printf(addDCond(Cond(w) << 1));
	else printf(subDCond(Cond(w) << 1));
	if (Nu(w)) printf(",N");
	printf("\t%%r%d,%%r%d,",Rsa(w),Rsb(w));
	psymoff(tgtofs, ISYM, "");
	return(YES);
}

/* Compare/Add and Branch Immediate instructions */
BOOL cbiDasm(struct inst *i,OFS ofs)
{
	FAST OFS tgtofs = ofs + 8 + Cbdisp(w);
	uint rmdr;
	char *buf, *s;

	if (Match("MOVIB")) printf(edDCond(Cond(w)));
	else if (Match("ADDIB")) printf(addDCond(Cond(w) << 1));
	else printf(subDCond(Cond(w) << 1));
	if (Nu(w)) printf(",N");
	printf("\t%d,%%r%d,",Ima5(w),Rsb(w));
	psymoff(tgtofs, ISYM, "");
	return(YES);
}

/* Branch on Bit instructions */
BOOL bbDasm(struct inst *i,OFS ofs)
{
	FAST OFS tgtofs = ofs + 8 + Cbdisp(w);
	uint rmdr;
	char *buf, *s;

	printf(edDCond(Cond(w)));
	if (Nu(w)) printf(",N");
	if (Match("BVB")) printf("\t%%r%d,",Rta(w));
	else printf("\t%%r%d,%d,",Rsa(w),Imb5(w));
	psymoff(tgtofs, ISYM, "");
	return(YES);
}

/* Arithmetic instructions */
BOOL ariDasm(struct inst *i)
{
	if (Match("OR") && Rsb(w) == 0 && Cond4(w) == NEV) {
		if (Rsa(w) == 0 && Rtc(w) == 0)
			printf("NOP");
		else
			printf("COPY\t%%r%d,%%r%d",Rsa(w),Rtc(w));
		return(YES);
	}
	printf(i->mnem);
	printf(subDCond(Cond4(w)));
	printf("\t%%r%d,%%r%d,%%r%d",Rsa(w),Rsb(w),Rtc(w));
	return(YES);
}

/* Add instructions */
BOOL addDasm(void)
{
	printf(addDCond(Cond4(w)));
	printf("\t%%r%d,%%r%d,%%r%d",Rsa(w),Rsb(w),Rtc(w));
	return(YES);
}


/* Unit instructions */
BOOL unitDasm(struct inst *i)
{
	printf(unitDCond(Cond4(w)));
	if (Match("DCOR") || Match("IDCOR")) printf("\t%%r%d,%%r%d",Rsb(w),Rtc(w));
	else printf("\t%%r%d,%%r%d,%%r%d",Rsa(w),Rsb(w),Rtc(w));
	return(YES);
}

/* Immediate Arithmetic instructions */
BOOL iaDasm(struct inst *i)
{
	if (Match("ADDI")) printf("%s\t%d,%%r%d,%%r%d",addDCond(Cond4(w)),Im11(w),Rsb(w),Rta(w));
	else printf("%s\t%d,%%r%d,%%r%d",subDCond(Cond4(w)),Im11(w),Rsb(w),Rta(w));
	return(YES);
}

/* Shift double instructions */
BOOL shdDasm(struct inst *i)
{
	if (Match("VSHD"))
	    printf("%s\t%%r%d,%%r%d,%%r%d",edDCond(Cond(w)), Rsa(w),Rsb(w),Rtc(w));
	else printf("%s\t%%r%d,%%r%d,%d,%%r%d",edDCond(Cond(w)),Rsa(w),Rsb(w),31-Imd5(w),Rtc(w));
	return(YES);
}

/* Extract instructions */
BOOL extrDasm(void)
{
	printf("%s\t%%r%d,%d,%d,%%r%d",edDCond(Cond(w)),Rsb(w),Imd5(w),32 - Rsc(w),Rta(w));
	return(YES);
}


/* Variable extract instructions */
BOOL vextrDasm(void)
{
	printf("%s\t%%r%d,%d,%%r%d",edDCond(Cond(w)),Rsb(w),32 - Rsc(w),Rta(w));
	return(YES);
}


/* Deposit instructions */
BOOL depDasm(void)
{
	printf("%s\t%%r%d,%d,%d,%%r%d",edDCond(Cond(w)),Rsa(w),31 - Imd5(w),32 - Rsc(w),Rtb(w));
	return(YES);
}


/* Variable deposit instructions */
BOOL vdepDasm(void)
{
	printf("%s\t%%r%d,%d,%%r%d",edDCond(Cond(w)),Rsa(w),32 - Rsc(w),Rtb(w));
	return(YES);
}


/* Deposit Immediate instructions */
BOOL depiDasm(void)
{
	printf("%s\t%d,%d,%d,%%r%d",edDCond(Cond(w)),Ima5(w),31 - Imd5(w),32 - Imc5A(w),Rtb(w));
	return(YES);
}

/* Variable Deposit Immediate instructions */
BOOL vdepiDasm(void)
{
	printf("%s\t%d,%d,%%r%d",edDCond(Cond(w)),Ima5(w),32-Imc5A(w),Rtb(w));
	return(YES);
}

/* System control operations */
BOOL scDasm(struct inst *i)
{
	if (Match("MTCTL")) {
		if (Rtb(w) == 11)
			printf("MTSAR\t%%r%d",Rsa(w));
		else
			printf("MTCTL\t%%r%d,%%cr%d",Rsa(w),Rtb(w));
		return(YES);
	}
	printf(i->mnem);
	if (Match("SSM") || Match("RSM")) printf("\t%d,%%r%d",Ima5A(w),Rtc(w));
	else if (Match("MTSM")) printf("\t%%r%d",Rsa(w));
	else if (Match("LDPRID")) printf("\t%%r%d",Rtc(w));
	else if (Match("MTSP")) printf("\t%%r%d,%%sr%d",Rsa(w),Sr(w));
	else if (Match("MFSP")) printf("\t%%sr%d,%%r%d",Sr(w),Rtc(w));
	else if (Match("MFCTL")) printf("\t%%cr%d,%%r%d",Rsb(w),Rtc(w));
	else if (Match("LDSID")) {
		if (Dss(w))
			printf("\t(%%sr%d,%%r%d),%%r%d",Dss(w),Rsb(w),Rtc(w));
		else
			printf("\t(%%r%d),%%r%d",Rsb(w),Rtc(w));
	} else {
		printf("?????");
		return(NO);
	}
	return(YES);
}

/* Instruction cache/tlb control instructions */
BOOL mmgtDasm(struct inst *i)
{
	if (Match("PROBE")) {
		if (ProbeI(w)) {
			if (Dss(w))
				printf("I\t(%%sr%d,%%r%d),%d,%%r%d",Dss(w),Rsb(w),Rsa(w),Rtc(w));
			else
				printf("I\t(%%r%d),%d,%%r%d",Rsb(w),Rsa(w),Rtc(w));
		} else {
			if (Dss(w))
				printf("\t(%%sr%d,%%r%d),%%r%d,%%r%d",Dss(w),Rsb(w),Rsa(w),Rtc(w));
			else
				printf("\t(%%r%d),%%r%d,%%r%d",Rsb(w),Rsa(w),Rtc(w));
		}
	}
	else if (Match("LHA") || Match("LPA")) {
		if (Modify(w)) printf(",M");
		if (Dss(w))
			printf("\t%%r%d(%%sr%d,%%r%d),%%r%d",Rsa(w),Dss(w),Rsb(w),Rtc(w));
		else
			printf("\t%%r%d(%%r%d),%%r%d",Rsa(w),Rsb(w),Rtc(w));
	}
	else if (Match("PDTLB") || Match("PDC") || Match("FDC")) {
		if (Modify(w)) printf(",M");
		if (Dss(w))
			printf("\t%%r%d(%%sr%d,%%r%d)",Rsa(w),Dss(w),Rsb(w));
		else
			printf("\t%%r%d(%%r%d)",Rsa(w),Rsb(w));
	}
	else if (Match("PITLB") || Match("FIC")) {
		if (Modify(w)) printf(",M");
		printf("\t%%r%d(%%sr%d,%%r%d)",Rsa(w),Sr(w),Rsb(w));
	}
	else if (Match("IDTLB")) {
		if (Dss(w))
			printf("\t%%r%d,(%%sr%d,%%r%d)",Rsa(w),Dss(w),Rsb(w));
		else
			printf("\t%%r%d,(%%r%d)",Rsa(w),Rsb(w));
	}
	else if (Match("IITLB"))
		printf("\t%%r%d,(%%sr%d,%%r%d)",Rsa(w),Sr(w),Rsb(w));
	else {
		printf("?????");
		return(NO);
	}
	return(YES);
}

/* break instruction */
BOOL brkDasm(void)
{
	printf("\t%d,%d",Bi1(w),Bi2(w));
	return(YES);
}

BOOL floatDasm(void)
{
	FAST uint op1, r1, fmt, t;
	uint op2, r2, dfmt;
	char *p, **st, **dt;

	op1 = CoprExt1(w);
	op2 = CoprExt2(w);
	fmt = (op1 >> 2) & 3;		/* get precision of source  */
	st = (fmt & 1) ? fdreg : fsreg;	/* get source register type */
	/*
	 * get first (or only) source register
	 * (independent of class)
	 */
	r1 = (op1 >> 11) & 0x3e;
	if ((fmt & 1) == 0 && (Uid(w) & 2))
		r1++;

	if (op1 & 2) {				/* class 2 or 3 */
		/*
		 * get second source register
		 */
		r2 = (op1 >> 6) & 0x3e;
		if (fmt == 2)
			r2++;

		if ((op1 & 1) == 0) {		/* class 2 */
			/* Opclass 2: 2 sources, no destination */
			switch((op1 >> 4) & 7) {
			    case 0: p = "CMP"; break;
			    default: printf(fcoprUndef); return(NO);
			}
			printf("%s,%s",p,fmtStrTbl[fmt]);
			printf(",%s\t%%f%s,%%f%s",condStrTbl[op2],st[r1],st[r2]);
			return(YES);
		}
		/*
		 * get target register (class 3)
		 */
		t = (op2 << 1);
		if ((fmt & 1) == 0 && (Uid(w) & 1))
			t++;
		/* Opclass 3: 2 sources, 1 destination */
		switch((op1 >> 4) & 7) {
		    case 0: p = "ADD"; break;
		    case 1: p = "SUB"; break;
		    case 2: p = (Fpi(w)) ? "MPYI" : "MPY"; break;
		    case 3: p = "DIV"; break;
		    case 4: p = "REM"; break;
		    default: printf(fcoprUndef); return(NO);
		}
		printf("%s,%s",p,fmtStrTbl[fmt]);
		printf("\t%%f%s,%%f%s,%%f%s",st[r1],st[r2],st[t]);
	} else if (op1 & 1) {			/* class 1 */
		dfmt = (op1 >> 4) & 3;
		dt = (dfmt & 1) ? fdreg : fsreg;
		/*
		 * get target register
		 */
		t = (op2 << 1);
		if ((dfmt & 1) == 0 && (Uid(w) & 1))
			t++;
		/* Opclass 1: 1 source, 1 destination conversions */
		switch((op1 >> 6) & 3) {
		    case 0: p = "FF"; break;
		    case 1: p = "XF"; break;
		    case 2: p = "FX"; break;
		    case 3: p = "FXT"; break;
		}
		printf("%s,%s",p,fmtStrTbl[fmt]);
		printf(",%s\t%%f%s,%%f%s",fmtStrTbl[dfmt],st[r1],dt[t]);
	} else {				/* class 0 */
		/*
		 * get target register
		 */
		t = (op2 << 1);
		if ((fmt & 1) == 0 && (Uid(w) & 1))
			t++;
		/* Opclass 0: 1 source, 1 destination */
		switch((op1 >> 4) & 7) {
		    case 1: p = "RSQRT"; break;
		    case 2: p = "CPY"; break;
		    case 3: p = "ABS"; break;
		    case 4: p = "SQRT"; break;
		    case 5: p = "RND"; break;
		    default: printf(fcoprUndef); return(NO);
		}
		printf("%s,%s",p,fmtStrTbl[fmt]);
		printf("\t%%f%s,%%f%s",st[r1],st[t]);
	}
	return(YES);
}

BOOL fcoprDasm(uint op1,uint op2)
{
	FAST uint r1, r2, t, fmt, dfmt;
	FAST char *p;

	if (AstNu(w) && op1 == ((1<<4) | 2)) {
		if (op2 == 0) {
			printf("FTEST"); return(YES);
		}
		if (op2 == 1) {
			printf("FTEST,ACC"); return(YES);
		}
		if (op2 == 2) {
			printf("FTEST,REJ"); return(YES);
		}
		return(NO);
	} else if (0 == op1 && 0 == op2) {
		printf("FCOPR IDENTIFY"); return(YES);
	}
	switch(op1 & 3) {
	    case 0:
		/* Opclass 0: 1 source, 1 destination */
		r1 = (op1 >> 12) & 0x1f; t = op2; fmt = (op1 >> 2) & 3;
		switch((op1 >> 4) & 7) {
		    case 1: p = "FRSQRT"; break;
		    case 2: p = "FCPY"; break;
		    case 3: p = "FABS"; break;
		    case 4: p = "FSQRT"; break;
		    case 5: p = "FRND"; break;
		    default: printf(fcoprUndef); return(NO);
		}
		printf("%s,%s\t%%fr%d,%%fr%d",p,fmtStrTbl[fmt],r1,t);
		return(YES);
	    case 1:
		/* Opclass 1: 1 source, 1 destination conversions */
		r1 = (op1 >> 12) & 0x1f; t = op2;
		fmt = (op1 >> 2) & 3; dfmt = (op1 >> 4) & 3;
		switch((op1 >> 6) & 3) {
		    case 0: p = "FCNVFF"; break;
		    case 1: p = "FCNVXF"; break;
		    case 2: p = "FCNVFX"; break;
		    case 3: p = "FCNVFXT"; break;
		}
		printf("%s,%s,%s\t%%fr%d,%%fr%d",p,fmtStrTbl[fmt],fmtStrTbl[dfmt],r1,t);
		return(YES);
	    case 2:
		/* Opclass 2: 2 sources, no destination */
		r1 = (op1 >> 12) & 0x1f; r2 = (op1 >> 7) & 0x1f;
		fmt = (op1 >> 2) & 3;
		switch((op1 >> 4) & 7) {
		    case 0: p = "FCMP"; break;
		    default: printf(fcoprUndef); return(NO);
		}
		printf("%s,%s,%s\t%%fr%d,%%fr%d",p,fmtStrTbl[fmt],condStrTbl[op2],r1,r2);
		return(YES);
	    case 3:
		/* Opclass 3: 2 sources, 1 destination */
		r1 = (op1 >> 12) & 0x1f; r2 = (op1 >> 7) & 0x1f; t = op2;
		fmt = (op1 >> 2) & 3;
		switch((op1 >> 4) & 7) {
		    case 0: p = "FADD"; break;
		    case 1: p = "FSUB"; break;
		    case 2: p = "FMPY"; break;
		    case 3: p = "FDIV"; break;
		    case 4: p = "FREM"; break;
		    default: printf(fcoprUndef); return(NO);
		}
		printf("%s,%s\t%%fr%d,%%fr%d,%%fr%d",p,fmtStrTbl[fmt],r1,r2,t);
		return(YES);
	    default: printf(fcoprUndef); return(NO);
	}
	return(YES);
}

BOOL coprDasm(struct inst *i)
{
	FAST uint uid = Uid(w);
	FAST BOOL load = NO;
	FAST char *pfx = uid > 1 ? "C" : "F";
	FAST char **regt = fsreg;

	if (Match("COPR")) {
		if (uid) {
			printf("COPR,%d,0x%x",uid,CoprExt(w));
			if (AstNu(w)) printf(",N");
			return(YES);
		}
		return fcoprDasm(CoprExt1(w),CoprExt2(w));
	}
	if (Match("CLDD")) { regt = fdreg; load = YES; printf("%sLDD",pfx); }
	else if (Match("CLDW")) { load = YES; printf("%sLDW",pfx); }
	else if (Match("CSTD")) { regt = fdreg; printf("%sSTD",pfx); }
	else if (Match("CSTW")) printf("%sSTW",pfx);
	else { printf("COPR???"); return(NO); }
	if (ShortDisp(w)) {
		printf("S");
		if (AstNu(w)) {
			if (ModBefore(w)) printf(",MB");
			else printf(",MA");
		}
	}
	else {
		printf("X");
		if (AstNu(w)) {
			if (IndxShft(w)) printf(",SM");
			else printf(",M");
		}
		else if (IndxShft(w)) printf(",S");
	}
	switch (CacheCtrl(w)) {
		case NOACTION:	break;
		case STACKREF:	printf(",C"); break;
		case SEQPASS:	printf(",Q"); break;
		case PREFETCH:	printf(",P"); break;
	}
	if (load) {
		if (ShortDisp(w))
			printf("\t%d",Ima5(w));
		else
			printf("\t%%r%d",Rsa(w));
		if (Dss(w))
			printf("(%%sr%d,%%r%d),%%f%s",Dss(w),Rsb(w),regt[(Rtc(w)<<1)+(uid&1)]);
		else
			printf("(%%r%d),%%f%s",Rsb(w),regt[(Rtc(w)<<1)+(uid&1)]);
	} else {
		if (ShortDisp(w))
			printf("\t%%f%s,%d",regt[(Rsc(w)<<1)+(uid&1)],Ima5(w));
		else
			printf("\t%%f%s,%%r%d",regt[(Rsc(w)<<1)+(uid&1)],Rta(w));
		if (Dss(w))
			printf("(%%sr%d,%%r%d)",Dss(w),Rsb(w));
		else
			printf("(%%r%d)",Rsb(w));
	}
	return(YES);
}

BOOL lpkDasm(void)
{
	/*
	 * Floating point STore Quad
	 * Short or Indexed
	 */
	if (ShortDisp(w)) {
		if (Modify(w)) {
			if (ModBefore(w)) printf(",MB");
			else printf(",MA");
		}
	} else {
		if (Modify(w)) {
			if (IndxShft(w)) printf(",SM");
			else printf(",M");
		}
		else if (IndxShft(w)) printf(",S");
	}
	switch (CacheCtrl(w)) {
		case NOACTION:	break;
		case STACKREF:	printf(",C"); break;
		case SEQPASS:	printf(",Q"); break;
		case PREFETCH:	printf(",P"); break;
	}
	if (ShortDisp(w))
		printf("\t%%fr%d,%d",Rsc(w),Ima5(w));
	else
		printf("\t%%fr%d,%%r%d",Rsc(w),Rta(w));
	if (Dss(w))
		printf("(%%sr%d,%%r%d)",Dss(w),Rsb(w));
	else
		printf("(%%r%d)",Rsb(w));
	return(YES);
}

BOOL diagDasm(struct inst *i)
{
	if (0x0b0 == BitfR(w,19,8,_b198)) {	/* mtcpu */
		printf("MTCPU\t%%r%d,%%dr%d", Rsa(w), Rtb(w));
		return(YES);
	}
	if (0x0d0 == BitfR(w,19,8,_b198)) {	/* mfcpu */
		printf("MFCPU\t%%dr%d,%%r%d", Rsb(w), Rta(w));
		return(YES);
	}

	printf(i->mnem);
	if (Match("DIAG")) {
		printf("\t0x%X",w & 0x03ffffff);
		return(YES);
	}
	printf("?????");
	return(NO);
}

BOOL fmpysubDasm(void)
{
	char	**t = SinglePrec(w) ? fsreg : fdreg;

	printf((SinglePrec(w)) ? "SUB,SGL" : "SUB,DBL");
	printf("\t%%f%s,%%f%s,%%f%s,%%f%s,%%f%s",
		t[Ms1(w)], t[Ms2(w)], t[Mt(w)], t[As(w)], t[Ad(w)]);
	return(YES);
}

BOOL fmpyaddDasm(void)
{
	char	**t = SinglePrec(w) ? fsreg : fdreg;

	if (Rsd(w) == 0) {
		printf((SinglePrec(w)) ? "CFXT,SGL" : "CFXT,DBL");
		printf("\t%%f%s,%%f%s,%%f%s,%%f%s",
			t[Ms1(w)], t[Ms2(w)], t[Mt(w)], t[Ad(w)]);
	} else {
		printf((SinglePrec(w)) ? "ADD,SGL" : "ADD,DBL");
		printf("\t%%f%s,%%f%s,%%f%s,%%f%s,%%f%s",
			t[Ms1(w)], t[Ms2(w)], t[Mt(w)], t[As(w)], t[Ad(w)]);
	}
	return(YES);
}

/*---------------------------------------------------------------------------
 * instruction$Disassemble - Print a disassembled version of the passed word
 *  to standard output.
 *---------------------------------------------------------------------------*/
iDasm(FAST WORD instruct, OFS ofs)
{
	FAST struct inst *i;
	FAST struct majoropcode *m;
	FAST uint ext;
	BOOL rv;

	w = instruct;
	m = &majopcs[Opcode(w)];
	ext = OpExt(w, m);
	if (ext > m->maxsubop) {
		printf(illeg.mnem);
		return(NO);
	}
	i = m->subops[ext];
	if      (i->dasmfcn == coprDasm) rv = coprDasm(i);
	else if (i->dasmfcn == diagDasm) rv = diagDasm(i);
	else if (i->dasmfcn == ariDasm)  rv = ariDasm(i);
	else if (i->dasmfcn == scDasm)   rv = scDasm(i);
	else if (i->dasmfcn == ldDasm)   rv = ldDasm(i);
	else {
		printf(i->mnem);
		if (i->dasmfcn == mmgtDasm) rv = mmgtDasm(i);
		else if (i->dasmfcn) {
			VOID	(*f)(struct inst *,OFS,WORD) = i->dasmfcn;
			rv = (*f)(i,ofs,w);
		}
		else if (i->mnem[0] == '?') rv = NO;
		else                        rv = YES;
	}
	return(sizeof(instruct));
}

int print_instruction(inst,dot)
ulong inst;
unsigned dot;
{
	if (!unasm_inited)
		iExInit();
	unasm_inited = TRUE;

	return(iDasm((WORD)inst, dot));
}
