/*	$CHeader: tgdb.c 1.4 93/04/26 17:27:51 $ */

/*
 * Mach Operating System
 * Copyright (c) 1993, CONVEX Computer Corporation.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CONVEX ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION,
 * AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 */
/*
 * Machine dependent stub for remote gdb for user-level tasks.
 *
 * Known problem areas are marked with "FIXME".
 */

#include <kern/assert.h>
#include <kern/thread.h>
#include <mach/port.h>
#include <mach/thread_info.h>
#include <mach/thread_status.h>
#include <mach/vm_param.h>
#include <machine/psl.h>		/* for PSW_SS */

/*
 * Use the external versions of the kernel interface routines.
 */
#include <mach/mach_user_internal.h>
#include <mach/mach_port_internal.h>
#include <mach/mach_host_internal.h>

void
tgdb_net_init()
{
	pc586init(0);
}

void
tgdb_net_xmt(char *buffer, int length)
{
	pc586debug_xmt(0, buffer, length);
}	

void
tgdb_get_registers(mach_port_t thread, char *buffer, unsigned int *size)
{
	struct hp800_saved_state *ss = (struct hp800_saved_state *) buffer;
	
	assert(size >= sizeof(struct hp800_saved_state));

	bzero(buffer, *size);
	if (thread_abort(thread) != KERN_SUCCESS) {
		printf("tgdb: can't abort thread\n");
		return;
	}
	if (thread_get_state(thread, HP800_THREAD_STATE,
			     (thread_state_t) buffer, size) != KERN_SUCCESS) {
		printf("tgdb: can't get thread state\n");
		return;
	}

	ss->iioq_head &= ~3;
	ss->iioq_tail &= ~3;
}

void
tgdb_set_registers(mach_port_t thread, char *buffer, unsigned int size)
{
	assert(size >= sizeof(struct hp800_saved_state));

	thread_abort(thread);
	thread_set_state(thread, HP800_THREAD_STATE, (thread_state_t) buffer,
			 size);
}

/*
 * Set the trace bit in the thread's PSW.  This will cause it to
 * execute a single user-mode instruction, then raise a breakpoint
 * exception.
 */
void
tgdb_set_trace(mach_port_t thread)
{
	struct hp800_saved_state ss;
	unsigned int size = sizeof(ss);

	tgdb_get_registers(thread, (char *) &ss, &size);
	ss.cr22 |= PSW_SS;
	tgdb_set_registers(thread, (char *) &ss, size);
}
    
/*
 * Suspend thread, retrieve registers, and clear the trace bit in the
 * thread's PSW.
 */
void
tgdb_clear_trace(mach_port_t thread)
{
	struct hp800_saved_state ss;
	unsigned int size = sizeof(ss);

	if (thread_suspend(thread) != KERN_SUCCESS) {
		printf("tgdb: can't suspend thread\n");
		return;
	}
	tgdb_get_registers(thread, (char *) &ss, &size);
	ss.cr22 &= ~PSW_SS;
	tgdb_set_registers(thread, (char *) &ss, size);
}
    
/*
 * Retrieve basic information about each thread in the task.  This will
 * be simply formatted and printed for the remote gdb user.
 */
void
tgdb_basic_thread_info(mach_port_t thread, unsigned int *buffer)
{
	struct thread_basic_info tbi;
	unsigned int size = sizeof(tbi);
	struct hp800_saved_state ss;
	unsigned int ss_size = sizeof(ss);
	
	tgdb_get_registers(thread, (char *) &ss, &ss_size);
	thread_info(thread, THREAD_BASIC_INFO, &tbi, &size);
	*buffer++ = tbi.run_state;
	*buffer++ = tbi.flags;
	*buffer++ = tbi.suspend_count;
	*buffer++ = ss.iioq_head;
}
