/*
 * Mach Operating System
 * Copyright (c) 1990,1989 Carnegie Mellon University
 * Copyright (c) 1990,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: som.c 1.12 94/12/14$
 */

/*
 * hppa-specific routines for loading SOM executable files.
 */

#include <mach/vm_prot.h>
#include <mach/exec/exec.h>
#include <mach/machine/exec/som.h>

#define round_to_word(x) (((x) + sizeof(int)) & ~(sizeof(int) - 1))

/*
 * Machine-dependent portions of execve() for the hp800
 */

int exec_load_som(int (*read)(void *handle, vm_offset_t file_ofs, void *buf, vm_size_t size),
		  int (*read_exec)(void *handle, vm_offset_t file_ofs, vm_size_t file_size,
		   		   vm_offset_t mem_addr, vm_size_t mem_size, vm_prot_t mem_prot),
		  void *handle,
		  vm_offset_t *out_entry)
{
	som_exec	x;
	register int	result;
	vm_prot_t	text_prot;

	/*
	 * Read in all the necessary file headers at once.
	 */
	result = read(handle, 0, &x, sizeof(x));
	if (result)
		return (result);
	/* XXX should just return EX_NOT_EXECUTABLE if too short */

	/*
	 * We load 800 BSD images or 700 OSF SOM images
	 */
	if ((x.fhdr.system_id != MID_HP800)
	    && (x.fhdr.system_id != MID_HPUX))
		return (EX_NOT_EXECUTABLE);

	/*
	 * see if this is an OSF SOM image
	 */
	if (x.fhdr.system_id == MID_HPUX) {
		if (x.ehdr.som_auxhdr.type != 0xC)
			return (EX_NOT_EXECUTABLE);
		x.ehdr.exec_entry = x.ehdr.exec_flags;
	}

	switch ((int)x.fhdr.a_magic) {

	    case EXEC_MAGIC:
		text_prot = VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE;
		break;

	    case SHARE_MAGIC:
	    case DEMAND_MAGIC:
		if (x.ehdr.exec_tsize == 0) {
			return (EX_NOT_EXECUTABLE);
		}
		text_prot = VM_PROT_READ | VM_PROT_EXECUTE;
		break;

	    default:
		return (EX_NOT_EXECUTABLE);
	}

	if (x.ehdr.exec_tsize > 0)
	{
		result = read_exec(handle, x.ehdr.exec_tfile, x.ehdr.exec_tsize,
				   x.ehdr.exec_tmem, x.ehdr.exec_tsize, text_prot);
		if (result)
			return (result);
	}

	if (x.ehdr.exec_dsize + x.ehdr.exec_bsize > 0)
	{
		result = read_exec(handle, x.ehdr.exec_dfile, x.ehdr.exec_dsize,
				   x.ehdr.exec_dmem, x.ehdr.exec_dsize + x.ehdr.exec_bsize,
				   VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
		if (result)
			return (result);
	}

	*out_entry = x.ehdr.exec_entry;

	/*
	 * We'll use entry_2 to pass the default dp
	 */

	/*XXX lp->entry_2 = x.fhdr.presumed_dp; */

	return 0;
}

